"""Metadata module for managing Key Rings."""


def __init__(hub):
    hub.metadata.gcp.cloudkms.projects.locations.key_rings.PATH = (
        "projects/{project_id}/locations/{location_id}/keyRings/{key_ring_id}"
    )



import json

import yaml
from yaml import SafeDumper


class ResultsCollector:
    @property
    def findings(self):
        return self._findings

    @property
    def base_revision(self):
        return self._base_revision

    @property
    def current_revision(self):
        return self._current_revision

    def __init__(self, current_revision=None, base_revision=None):
        self._findings = []
        self._current_revision = current_revision
        self._base_revision = base_revision

    def has_breaking_finding(self):
        return (
            len([finding for finding in self.findings if finding.get("is_breaking")])
            > 0
        )

    def revisions(self, current_revision, base_revision):
        self._current_revision = current_revision
        self._base_revision = base_revision
        return self

    def add_breaking(self, schema_path, description):
        self._findings.append(
            {
                "schema_path": schema_path,
                "description": description,
                "is_breaking": True,
            }
        )

    def add_non_breaking(self, schema_path, description):
        self._findings.append(
            {
                "schema_path": schema_path,
                "description": description,
                "is_breaking": False,
            }
        )

    def report_as_json(self):
        report = {
            "base_revision": self.base_revision,
            "current_revision": self.current_revision,
            "findings": self.findings,
        }
        return json.dumps(report, indent=4)

    def report_as_yaml(self):
        report = {
            "base_revision": self.base_revision,
            "current_revision": self.current_revision,
            "findings": self.findings,
        }
        return yaml.dump(report, Dumper=SafeDumper)



from dataclasses import dataclass


@dataclass
class Finding:
    schema_path: str
    description: str
    is_breaking: bool

    def __str__(self):
        breaking_text = "breaking" if self.is_breaking else "non-breaking"
        return f"{self._schema_path} - {self.description} (a {breaking_text} change)"

    @classmethod
    def breaking(cls, schema_path, description):
        return Finding(
            schema_path=schema_path, description=description, is_breaking=True
        )

    @classmethod
    def non_breaking(cls, schema_path, description):
        return Finding(
            schema_path=schema_path, description=description, is_breaking=False
        )


# class EnhancedJSONEncoder(json.JSONEncoder):
#     def default(self, o):
#         if dataclasses.is_dataclass(o):
#             return dataclasses.asdict(o)
#         return super().default(o)



"""Exec module for managing Disks."""
__func_alias__ = {"list_": "list"}

from dataclasses import make_dataclass, field
from typing import Dict, Any, List

from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext


async def list_(
    hub,
    ctx,
    project: str = None,
    zone: str = None,
    region: str = None,
    filter_: (str, "alias=filter") = None,
    order_by: str = None,
):
    r"""Retrieves a list of persistent disks.

    Args:
        project(str, Optional):
            Project ID for this request.

        zone(str, Optional):
            The name of the zone for this request.

        region(str, Optional):
            The name of the region for this request.

        filter(str, Optional):
            A filter expression that filters resources listed in the response. Most Compute resources support two types of filter expressions: expressions that support regular expressions and expressions that follow API improvement proposal AIP-160. If you want to use AIP-160, your expression must specify the field name, an operator, and the value that you want to use for filtering. The value must be a string, a number, or a boolean. The operator must be either `=`, `!=`, `>`, `<`, `<=`, `>=` or `:`. For example, if you are filtering Compute Engine instances, you can exclude instances named `example-instance` by specifying `name != example-instance`. The `:` operator can be used with string fields to match substrings. For non-string fields it is equivalent to the `=` operator. The `:*` comparison can be used to test whether a key has been defined. For example, to find all objects with `owner` label use: ``` labels.owner:* ``` You can also filter nested fields. For example, you could specify `scheduling.automaticRestart = false` to include instances only if they are not scheduled for automatic restarts. You can use filtering on nested fields to filter based on resource labels. To filter on multiple expressions, provide each separate expression within parentheses. For example: ``` (scheduling.automaticRestart = true) (cpuPlatform = \"Intel Skylake\") ``` By default, each expression is an `AND` expression. However, you can include `AND` and `OR` expressions explicitly. For example: ``` (cpuPlatform = \"Intel Skylake\") OR (cpuPlatform = \"Intel Broadwell\") AND (scheduling.automaticRestart = true) ``` If you want to use a regular expression, use the `eq` (equal) or `ne` (not equal) operator against a single un-parenthesized expression with or without quotes or against multiple parenthesized expressions. Examples: `fieldname eq unquoted literal` `fieldname eq 'single quoted literal'` `fieldname eq \"double quoted literal\"` `(fieldname1 eq literal) (fieldname2 ne \"literal\")` The literal value is interpreted as a regular expression using Google RE2 library syntax. The literal value must match the entire field. For example, to filter for instances that do not end with name "instance", you would use `name ne .*instance`.

        order_by(str, Optional):
            Sorts list results by a certain order. By default, results are returned in alphanumerical order based on the resource name. You can also sort results in descending order based on the creation timestamp using `orderBy=\"creationTimestamp desc\"`. This sorts results based on the `creationTimestamp` field in reverse chronological order (newest result first). Use this to sort resources like operations so that the newest operation is returned first. Currently, only sorting by `name` or `creationTimestamp desc` is supported.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.disk.list
              - kwargs:
                  project: project-name
                  zone: zone-name
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if zone:
        ret = await hub.exec.gcp_api.client.compute.disk.list(
            ctx,
            project=project,
            zone=zone,
            filter=filter_,
            orderBy=order_by,
        )
    elif region:
        ret = await hub.exec.gcp_api.client.compute.disk.list(
            ctx,
            project=project,
            region=region,
            filter=filter_,
            orderBy=order_by,
        )
    else:
        ret = await hub.exec.gcp_api.client.compute.disk.aggregatedList(
            ctx,
            project=project,
            filter=filter_,
            orderBy=order_by,
        )

    if not ret["result"]:
        result["comment"] += ret["comment"]
        result["result"] = False
        return result

    result["ret"] = ret["ret"].get("items", [])
    return result


async def get(
    hub,
    ctx,
    project: str = None,
    zone: str = None,
    region: str = None,
    name: str = None,
    resource_id: str = None,
):
    r"""Returns a specified persistent disk.

    Use an un-managed disk as a data-source. Supply one of the inputs as the filter.
    Gets a list of available persistent disks by making a list() request.

    Args:
        project(str, Optional):
            Project ID for this request. Defaults to None.

        zone(str, Optional):
            The name of the zone for this request. Defaults to None.

        region(str, Optional):
            The name of the region for this request. Defaults to None.

        name(str, Optional):
            Name of the persistent disk to return. Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.disk.get
              - kwargs:
                  name: disk-name
                  project: project-name
                  zone: zone-name
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if resource_id:
        ret = await hub.exec.gcp_api.client.compute.disk.get(
            ctx,
            resource_id=resource_id,
        )
    elif project and zone and name:
        ret = await hub.exec.gcp_api.client.compute.disk.get(
            ctx, project=project, zone=zone, disk=name
        )
    elif project and region and name:
        ret = await hub.exec.gcp_api.client.compute.disk.get(
            ctx, project=project, region=region, disk=name
        )
    else:
        result["result"] = False
        result["comment"] = [
            f"gcp.compute.disk#get(): {name} either resource_id or project, zone/region and name"
            f" should be specified."
        ]
        return result

    result["comment"] += ret["comment"]
    if not ret["result"]:
        result["result"] = False
        return result

    result["ret"] = ret["ret"]
    return result


async def create_snapshot(
    hub,
    ctx,
    name: str,
    project: str = None,
    zone: str = None,
    region: str = None,
    disk: str = None,
    resource_id: str = None,
    storage_locations: List[str] = None,
    location_hint: str = None,
    label_fingerprint: str = None,
    description: str = None,
    labels: Dict[str, Any] = None,
    source_disk_encryption_key: make_dataclass(
        "CustomerEncryptionKey",
        [
            ("kms_key_service_account", str, field(default=None)),
            ("sha256", str, field(default=None)),
            ("rsa_encrypted_key", str, field(default=None)),
            ("kms_key_name", str, field(default=None)),
            ("raw_key", str, field(default=None)),
        ],
    ) = None,
    source_disk: str = None,
    chain_name: str = None,
    snapshot_type: str = None,
    snapshot_encryption_key: make_dataclass(
        "CustomerEncryptionKey",
        [
            ("kms_key_service_account", str, field(default=None)),
            ("sha256", str, field(default=None)),
            ("rsa_encrypted_key", str, field(default=None)),
            ("kms_key_name", str, field(default=None)),
            ("raw_key", str, field(default=None)),
        ],
    ) = None,
    request_id: str = None,
):
    r"""Creates a snapshot of a specified persistent disk. For regular snapshot creation, consider using snapshots.insert instead, as that method supports more features, such as creating snapshots in a project different from the source disk project.

    Args:
        name(str):
            Name of the resource; provided by the client when the resource is created. The name must be 1-63 characters long, and comply with RFC1035. Specifically, the name must be 1-63 characters long and match the regular expression `[a-z]([-a-z0-9]*[a-z0-9])?` which means the first character must be a lowercase letter, and all following characters must be a dash, lowercase letter, or digit, except the last character, which cannot be a dash.",
                                "pattern": "[a-z](?:[-a-z0-9]{0,61}[a-z0-9])?

        project(str, Optional):
            Project ID for this request. Defaults to None.

        zone(str, Optional):
            The name of the zone for this request. Defaults to None.

        region(str, Optional):
            The name of the region for this request. Defaults to None.

        disk(str, Optional):
            Name of the persistent disk to snapshot.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000).

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.disk.create_snapshot
              - kwargs:
                  name: snapshot-name
                  disk: disk-name
                  project: project-name
                  zone: zone-name
    """
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    request_body = {
        "name": name,
        "storage_locations": storage_locations,
        "location_hint": location_hint,
        "label_fingerprint": label_fingerprint,
        "description": description,
        "labels": labels,
        "source_disk_encryption_key": source_disk_encryption_key,
        "source_disk": source_disk,
        "chain_name": chain_name,
        "snapshot_type": snapshot_type,
        "snapshot_encryption_key": snapshot_encryption_key,
    }

    execution_context = ExecutionContext(
        resource_type="compute.disk",
        method_name="createSnapshot",
        method_params={
            "ctx": ctx,
            "resource_id": resource_id,
            "body": request_body,
            "requestId": request_id,
            "project": project,
            "zone": zone,
            "region": region,
            "disk": disk,
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def create_from_snapshot(
    hub,
    ctx,
    source_snapshot: str,
    size_gb: str,
    type_: (str, "alias=type") = None,
    name: str = None,
    project: str = None,
    zone: str = None,
    region: str = None,
    resource_id: str = None,
    request_id: str = None,
):
    r"""Create a disk from a snapshot.

    If you backed up a boot or non-boot disk with a snapshot, you can create a new disk based on the snapshot.

    Restrictions:
        The new disk must be at least the same size as the original source disk for the snapshot. If you create a disk that is larger than the original source disk for the snapshot, you must resize the file system on that persistent disk to include the additional disk space. Depending on your operating system and file system type, you might need to use a different file system resizing tool. For more information, see your operating system documentation.

        You can create a new zonal disk from a given snapshot at most once every ten minutes. If you want to issue a burst of requests to snapshot your disks, you can issue at most 6 requests in 60 minutes. This limit does not apply when creating regional disks from a snapshot.

    Args:
        name(str):
            Name of the resource; provided by the client when the resource is created. The name must be 1-63 characters long, and comply with RFC1035. Specifically, the name must be 1-63 characters long and match the regular expression `[a-z]([-a-z0-9]*[a-z0-9])?` which means the first character must be a lowercase letter, and all following characters must be a dash, lowercase letter, or digit, except the last character, which cannot be a dash.",
                                "pattern": "[a-z](?:[-a-z0-9]{0,61}[a-z0-9])?

        project(str, Optional):
            Project ID for this request. Defaults to None.

        zone(str, Optional):
            The name of the zone for this request. Defaults to None.

        region(str, Optional):
            The name of the region for this request. Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000).

        source_snapshot(str):
            The source snapshot used to create this disk. You can provide this as a partial or full URL to the resource.
            For example, the following are valid values:
                - https://www.googleapis.com/compute/v1/projects/project /global/snapshots/snapshot
                - projects/project/global/snapshots/snapshot
                - global/snapshots/snapshot

        size_gb(str):
            Size, in GB, of the persistent disk.
            You can specify this field when creating a persistent disk using the sourceImage, sourceSnapshot,
            or sourceDisk parameter, or specify it alone to create an empty persistent disk.
            If you specify this field along with a source, the value of sizeGb must not
            be less than the size of the source. Acceptable values are 1 to 65536, inclusive.
            Defaults to None.

        type(str, Optional):
            URL of the disk type resource describing which disk type to use to create the disk.
            Provide this when creating the disk. For example: projects/project /zones/zone/diskTypes/pd-ssd.
            See Persistent disk types.
            Defaults to None.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.disk.create_from_snapshot
              - kwargs:
                  source_snapshot: snapshot_resource_id
                  name: restored-disk-name
                  project: project-name
                  zone: zone-name
                  size_gb: disk-size
    """
    request_body = {
        "type": type_,
        "name": name,
        "size_gb": size_gb,
        "zone": zone,
        "region": region,
        "source_snapshot": source_snapshot,
    }

    execution_context = ExecutionContext(
        resource_type="compute.disk",
        method_name="insert",
        method_params={
            "ctx": ctx,
            "resource_id": resource_id,
            "project": project,
            "zone": zone,
            "region": region,
            "body": request_body,
            "requestId": request_id,
        },
    )

    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)



"""Google Cloud Platform provider exceptions module.

Copyright (c) 2021 VMware, Inc. All Rights Reserved.
SPDX-License-Identifier: Apache-2.0
"""


class StateReturn(dict):
    """Convenience class to manage POP state returns."""

    def __init__(self, name=None, result=None, comment=None, old_obj={}, new_obj={}):
        """Initialize an object of this class.

        :param name: The name of the state (e.g. from a SLS file).
        :param result: True if the state call works, False otherwise.
        :param comment: Any relevant comments to the state execution.
        For example a 200 code from an HTTP call or a mere readable comment.
        In some cases, users of StateReturn may also store additional 'comment'
        type material, such as Operations for use in future checks on the state
        of the operation.
        :param old_obj: For state changes, the object's state
        (e.g., dict of values).
        prior to any state change request executions.
        :param new_obj: For state changes, the new object's state after any
        state change equest executions.
        """
        super().__init__(
            [("name", name), ("result", result), ("comment", comment), ("changes", {})]
        )
        self["changes"] = {"old": old_obj if old_obj else {}, "new": new_obj}



"""Exec module for managing ForwardingRule."""
from typing import Any
from typing import Dict

from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext
from idem_gcp.tool.gcp.generate.exec_param import ExecParam
from idem_gcp.tool.gcp.generate.scope import Scope

__func_alias__ = {"list_": "list"}


async def list_(
    hub,
    ctx,
    project: str = None,
    region: str = None,
    filter_: (str, "alias=filter") = None,
    order_by: str = None,
):
    r"""Retrieves the list of ForwardingRule resources available to the specified project.

    Args:
        project(str, Optional):
            Project ID for this request.

        region(str, Optional):
            The name of the region for this request.

        filter(str, Optional):
            A filter expression that filters resources listed in the response. Most Compute resources support two types of filter expressions: expressions that support regular expressions and expressions that follow API improvement proposal AIP-160. If you want to use AIP-160, your expression must specify the field name, an operator, and the value that you want to use for filtering. The value must be a string, a number, or a boolean. The operator must be either `=`, `!=`, `>`, `<`, `<=`, `>=` or `:`. For example, if you are filtering Compute Engine instances, you can exclude instances named `example-instance` by specifying `name != example-instance`. The `:` operator can be used with string fields to match substrings. For non-string fields it is equivalent to the `=` operator. The `:*` comparison can be used to test whether a key has been defined. For example, to find all objects with `owner` label use: ``` labels.owner:* ``` You can also filter nested fields. For example, you could specify `scheduling.automaticRestart = false` to include instances only if they are not scheduled for automatic restarts. You can use filtering on nested fields to filter based on resource labels. To filter on multiple expressions, provide each separate expression within parentheses. For example: ``` (scheduling.automaticRestart = true) (cpuPlatform = \"Intel Skylake\") ``` By default, each expression is an `AND` expression. However, you can include `AND` and `OR` expressions explicitly. For example: ``` (cpuPlatform = \"Intel Skylake\") OR (cpuPlatform = \"Intel Broadwell\") AND (scheduling.automaticRestart = true) ``` If you want to use a regular expression, use the `eq` (equal) or `ne` (not equal) operator against a single un-parenthesized expression with or without quotes or against multiple parenthesized expressions. Examples: `fieldname eq unquoted literal` `fieldname eq 'single quoted literal'` `fieldname eq \"double quoted literal\"` `(fieldname1 eq literal) (fieldname2 ne \"literal\")` The literal value is interpreted as a regular expression using Google RE2 library syntax. The literal value must match the entire field. For example, to filter for instances that do not end with name "instance", you would use `name ne .*instance`.

        order_by(str, Optional):
            Sorts list results by a certain order. By default, results are returned in alphanumerical order based on the resource name. You can also sort results in descending order based on the creation timestamp using `orderBy=\"creationTimestamp desc\"`. This sorts results based on the `creationTimestamp` field in reverse chronological order (newest result first). Use this to sort resources like operations so that the newest operation is returned first. Currently, only sorting by `name` or `creationTimestamp desc` is supported.

    Examples:
        random-name:
          exec.run:
          - path: gcp.compute.forwarding_rule.list
          - kwargs:
              project: project-name
    """
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    execution_context = ExecutionContext(
        resource_type="compute.forwarding_rule",
        method_name="list",
        method_params={
            "ctx": ctx,
            "project": project,
            "region": region,
            "filter": filter_,
            "order_by": order_by,
        },
        exec_params={
            ExecParam.SCOPED_FUNCTIONS: {
                Scope.REGIONAL: "list",
                Scope.GLOBAL: "aggregatedList",
            }
        },
    )

    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def get(
    hub,
    ctx,
    region: str = None,
    name: str = None,
    project: str = None,
    resource_id: str = None,
):
    r"""Returns the specified ForwardingRule resource. Gets a list of available forwarding rules by making a list() request.

    Args:
        project(str, Optional):
            Project ID for this request. Defaults to None.

        region(str, Optional):
            The name of the region for this request.

        name(str, Optional):
            Name of the resource to return. Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.forwarding_rule.get
              - kwargs:
                  project: project-name
                  name: forwarding-rule-name
    """
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    execution_context = ExecutionContext(
        resource_type="compute.forwarding_rule",
        method_name="get",
        method_params={
            "ctx": ctx,
            "resource_id": resource_id,
            "project": project,
            "region": region,
            "forwardingRule": name,
        },
    )

    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def set_target(
    hub,
    ctx,
    target: str,
    forwarding_rule: str = None,
    project: str = None,
    region: str = None,
    resource_id: str = None,
):
    r"""Changes target URL for the GlobalForwardingRule resource. The new target should be of the same type as the old target.

    Args:
        target(str):
            The URL of the target resource to receive the matched traffic. For regional forwarding rules, this target must be in the same region as the forwarding rule. For global forwarding rules, this target must be a global load balancing resource. The forwarded traffic must be of a type appropriate to the target object. - For load balancers, see the \"Target\" column in [Port specifications](https://cloud.google.com/load-balancing/docs/forwarding-rule-concepts#ip_address_specifications). - For Private Service Connect forwarding rules that forward traffic to Google APIs, provide the name of a supported Google API bundle: - vpc-sc - APIs that support VPC Service Controls. - all-apis - All supported Google APIs. - For Private Service Connect forwarding rules that forward traffic to managed services, the target must be a service attachment.

        forwarding_rule(str, Optional):
            Name of the ForwardingRule resource in which target is to be set.

        project(str, Optional):
            Project ID for this request. Defaults to None.

        region(str, Optional):
            The name of the region for this request.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.forwarding_rule.get
              - kwargs:
                  project: project-name
                  name: forwarding-rule-name
    """
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    execution_context = ExecutionContext(
        resource_type="compute.forwarding_rule",
        method_name="setTarget",
        method_params={
            "ctx": ctx,
            "resource_id": resource_id,
            "project": project,
            "region": region,
            "forwarding_rule": forwarding_rule,
            "body": {"target": target},
        },
    )

    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def set_labels(
    hub,
    ctx,
    labels: Dict[str, Any],
    label_fingerprint: str,
    resource_id: str = None,
    project: str = None,
    region: str = None,
    resource: str = None,
):
    r"""Sets the labels on a forwarding rule. To learn more about labels, read the Labeling Resources documentation.

    Args:
        labels(Dict[str, Any]):
            A list of labels to apply for this resource. Each label must comply with the requirements for labels. For example, \"webserver-frontend\": \"images\". A label value can also be empty (e.g. \"my-label\": \"\").

        label_fingerprint(str):
            The fingerprint of the previous set of labels for this resource, used to detect conflicts. The fingerprint is initially generated by Compute Engine and changes after every request to modify or update labels. You must always provide an up-to-date fingerprint hash when updating or changing labels, otherwise the request will fail with error 412 conditionNotMet. Make a get() request to the resource to get the latest fingerprint.

        resource_id(str, Optional):
            An identifier of the resource in the provider.

        project(str, Optional):
            Project ID for this request.

        region(str, Optional):
            The name of the region for this request.

        resource(str, Optional):
            Name or id of the resource for this request.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.forwarding_rule.set_labels
              - kwargs:
                  project: project-name
                  resource: forwarding-rule-name
                  labels:
                    label1: value1
                    label2: value2
                  label_fingerprint: IujiSXsfa3M=
    """
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    request_body = {
        "labels": labels,
        "label_fingerprint": label_fingerprint,
    }

    execution_context = ExecutionContext(
        resource_type="compute.forwarding_rule",
        method_name="setLabels",
        method_params={
            "ctx": ctx,
            "resource_id": resource_id,
            "project": project,
            "region": region,
            "resource": resource,
            "body": request_body,
        },
        exec_params={
            ExecParam.PATH_PARAM_ALIASES: {
                # Path of forwardingRule#setLabels() has {resource}; others, e.g. forwardingRule#get(), have {forwardingRule} instead
                "forwardingRule": "resource",
            }
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)



"""State module for managing ServiceAccountKeys."""
from typing import Any
from typing import Dict

__contracts__ = ["resource"]
RESOURCE_TYPE = "iam.projects.service_accounts.key"
RESOURCE_TYPE_FULL = "gcp.iam.projects.service_accounts.key"


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    private_key_type: str = None,
    key_algorithm: str = None,
    key_origin: str = None,
    key_type: str = None,
    valid_before_time: str = None,
    valid_after_time: str = None,
    service_account_id: str = None,
):
    """Create a service account key.

    Args:
        name(str, Optional): The resource name of the service account key.
        resource_id(str, Optional): An identifier of the resource in the provider. Defaults to None.
        service_account_id(str, Optional): Required on create. The account resource id used to create a service key
        private_key_type(str, Optional): The output format of the private key. The default value is `TYPE_GOOGLE_CREDENTIALS_FILE`, which is the Google Credentials File format.
        key_algorithm(str, Optional): Which type of key and algorithm to use for the key. The default is currently a 2K RSA key. However this may change in the future.
        key_origin(str, Optional): The key origin. One of "ORIGIN_UNSPECIFIED", "USER_PROVIDED", "GOOGLE_PROVIDED".
        key_type(str, Optional): The key type. One of "KEY_TYPE_UNSPECIFIED", "USER_MANAGED", "SYSTEM_MANAGED"
        valid_before_time(str, Optional): The key can be used before this timestamp. For system-managed key pairs, this timestamp is the end time for the private key signing operation. The public key could still be used for verification for a few hours after this time.
        valid_after_time(str, Optional): The key can be used after this timestamp.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              gcp.iam.projects.service_accounts.key.present:
              - private_key_type: TYPE_PKCS12_FILE
              - key_algorithm: KEY_ALG_RSA_2048
              - service_account_id: projects/project/serviceAccounts/test@test.iam.gserviceaccount.com
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    if resource_id:
        old_get_ret = await hub.exec.gcp.iam.projects.service_accounts.key.get(
            ctx, resource_id=resource_id
        )

        if not old_get_ret["result"] or (not old_get_ret["ret"] and ctx["rerun_data"]):
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            return result

        result["old_state"] = old_get_ret["ret"]

    if result["old_state"]:
        resource_body = {
            "name": resource_id,
            "key_algorithm": key_algorithm,
            "key_origin": key_origin,
            "key_type": key_type,
            "valid_after_time": valid_after_time,
            "valid_before_time": valid_before_time,
        }
        resource_body = {k: v for (k, v) in resource_body.items() if v is not None}

        # no update operation is applicable to service account keys; simply compare states
        resource_body["resource_id"] = resource_id
        changes = hub.tool.gcp.utils.compare_states(
            result["old_state"],
            resource_body,
            RESOURCE_TYPE,
        )

        if not changes:
            result["comment"].append(
                hub.tool.gcp.comment_utils.up_to_date_comment(RESOURCE_TYPE_FULL, name)
            )
            result["new_state"] = result["old_state"]
            return result

        result["result"] = False
        result["comment"].append(
            hub.tool.gcp.comment_utils.no_resource_update_comment(
                RESOURCE_TYPE_FULL, resource_id
            )
        )
        return result
    else:
        resource_body = {
            "private_key_type": private_key_type,
            "key_algorithm": key_algorithm,
        }
        resource_body = {k: v for (k, v) in resource_body.items() if v is not None}

        if ctx.get("test", False):
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    RESOURCE_TYPE_FULL, name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                resource_body
            )
            result["new_state"]["resource_id"] = resource_id
            return result

        # Create
        create_ret = (
            await hub.exec.gcp_api.client.iam.projects.service_accounts.key.create(
                ctx,
                name=service_account_id,
                body=resource_body,
            )
        )
        if not create_ret["result"] or not create_ret["ret"]:
            result["result"] = False
            if create_ret["comment"] and next(
                (
                    comment
                    for comment in create_ret["comment"]
                    if "alreadyExists" in comment
                ),
                None,
            ):
                result["comment"].append(
                    hub.tool.gcp.comment_utils.already_exists_comment(
                        RESOURCE_TYPE_FULL, name
                    )
                )
            else:
                result["comment"] += create_ret["comment"]
            return result

        result["comment"].append(
            hub.tool.gcp.comment_utils.create_comment(
                RESOURCE_TYPE_FULL, create_ret["ret"]["resource_id"]
            )
        )

        result["new_state"] = create_ret["ret"]

        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
):
    r"""Deletes a service account key.

    Args:
        name(str):
            The name of the resource

        resource_id(str, Optional):
            The resource_id of the resource

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

        resource_is_absent:
          gcp.iam.projects.service_accounts.key.absent
    """
    result = {
        "result": True,
        "old_state": ctx.get("old_state"),
        "new_state": None,
        "name": name,
        "comment": [],
    }

    if not resource_id:
        # we don't have enough information to know what to delete
        result["comment"].append(
            hub.tool.gcp.comment_utils.already_absent_comment(RESOURCE_TYPE_FULL, name)
        )
        return result

    get_ret = await hub.exec.gcp.iam.projects.service_accounts.key.get(
        ctx, resource_id=resource_id
    )

    if not get_ret["result"]:
        result["result"] = False
        result["comment"].append(
            hub.tool.gcp.comment_utils.resource_not_found_comment(
                RESOURCE_TYPE_FULL, resource_id
            )
        )
        result["comment"].extend(get_ret["comment"])
        return result

    if not get_ret["ret"]:
        result["comment"].append(
            hub.tool.gcp.comment_utils.already_absent_comment(RESOURCE_TYPE_FULL, name)
        )
        return result

    result["old_state"] = get_ret["ret"]

    if ctx["test"]:
        result["comment"].append(
            hub.tool.gcp.comment_utils.would_delete_comment(RESOURCE_TYPE_FULL, name)
        )
        return result

    del_ret = await hub.exec.gcp_api.client.iam.projects.service_accounts.key.delete(
        ctx, name=resource_id
    )

    if not del_ret["result"]:
        result["result"] = False
        result["comment"].extend(del_ret["comment"])
        return result

    result["comment"].append(
        hub.tool.gcp.comment_utils.delete_comment(RESOURCE_TYPE_FULL, name)
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves the list of available keys for a given service account.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.iam.projects.service_accounts.key
    """
    result = {}

    list_accounts_ret = await hub.exec.gcp.iam.projects.service_account.list(
        ctx, project=ctx.acct.project_id
    )

    if not list_accounts_ret["result"]:
        hub.log.debug(
            f"Could not describe {RESOURCE_TYPE_FULL} {list_accounts_ret['comment']}"
        )
        return result

    for resource in list_accounts_ret["ret"]:
        sa_resource_id = resource.get("resource_id")

        key_ret = await hub.exec.gcp.iam.projects.service_accounts.key.list(
            ctx, project=ctx.acct.project_id, sa_resource_id=sa_resource_id
        )

        if not key_ret["result"]:
            hub.log.debug(
                f"Could not describe {RESOURCE_TYPE_FULL} in {sa_resource_id}: {key_ret['comment']}"
            )
        else:
            for service_account_key in key_ret["ret"]:
                resource_id = service_account_key.get("resource_id")
                result[resource_id] = {
                    "gcp.iam.projects.service_accounts.key.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in service_account_key.items()
                    ]
                }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



# https://pop.readthedocs.io/en/latest/tutorial/quickstart.html#adding-configuration-data
# In this dictionary goes all the immutable values you want to show up under hub.OPT.idem_gcp
CONFIG = {
    "config": {
        "default": None,
        "help": "Load extra options from a configuration file onto hub.OPT.idem_gcp",
    },
    # pop-create options
    "services": {
        "default": [],
        "nargs": "*",
        "help": "The cloud services to target, defaults to all",
        "dyne": "pop_create",
    },
}

# The selected subcommand for your cli tool will show up under hub.SUBPARSER
# The value for a subcommand is a dictionary that will be passed as kwargs to argparse.ArgumentParser.add_subparsers
SUBCOMMANDS = {
    # "my_sub_command": {}
    "gcp": {
        "help": "Create idem_gcp state modules by parsing Google Cloud Client libraries",
        "dyne": "pop_create",
    },
}

# Include keys from the CONFIG dictionary that you want to expose on the cli
# The values for these keys are a dictionaries that will be passed as kwargs to argparse.ArgumentParser.add_option
CLI_CONFIG = {
    "config": {"options": ["-c"]},
    # "my_option1": {"subcommands": ["A list of subcommands that exclusively extend this option"]},
    # This option will be available under all subcommands and the root command
    # "my_option2": {"subcommands": ["_global_"]},
    "services": {
        "subcommands": ["gcp"],
        "dyne": "pop_create",
    },
}

# These are the namespaces that your project extends
# The hub will extend these keys with the modules listed in the values
DYNE = {
    "states": ["states"],
    "exec": ["exec"],
    "tool": ["tool"],
    "acct": ["acct"],
    "metadata": ["metadata"],
    "autogen": ["autogen"],
}



"""Exec module for managing ServiceAccountKeys."""
from typing import Callable

from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext

__func_alias__ = {"list_": "list"}
RESOURCE_TYPE = "iam.projects.service_accounts.key"
RESOURCE_TYPE_FULL = "gcp.iam.projects.service_accounts.key"


async def list_(hub, ctx, sa_resource_id: str, project: str = None):
    r"""Lists every ServiceAccountKey that belongs to a specific project.

    Args:
        sa_resource_id(str):
            Resource id of the service account following the pattern projects/{project}/serviceAccounts/{id}.
        project(str, Optional):
            The resource name of the project associated with the service accounts.
    """
    execution_context = ExecutionContext(
        resource_type="iam.projects.service_accounts.key",
        method_name="list",
        method_params={"ctx": ctx, "name": sa_resource_id},
    )

    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def get(
    hub,
    ctx,
    project: str = None,
    service_account_id: str = None,
    key_id: str = None,
    resource_id: str = None,
):
    r"""Returns the specified ServiceAccountKey resource.

    Args:
        project(str, Optional):
            Project ID for this request.
        service_account_id(str, Optional):
            Email or unique_id of the service account in GCP.
        key_id(str, Optional):
            Id of the service account key in GCP.
        resource_id(str, Optional):
            An identifier of the service account key in idem. Defaults to None.
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    if service_account_id and key_id:
        project = hub.tool.gcp.utils.get_project_from_account(ctx, project)
        resource_id = (
            f"projects/{project}/serviceAccounts/{service_account_id}/keys/{key_id}"
        )
    elif not resource_id:
        result["result"] = False
        result["comment"] = [
            f"gcp.iam.projects.service_accounts.key#get(): either service_account_id and key_id or resource_id"
            f" should be specified."
        ]
        return result

    execution_context = ExecutionContext(
        resource_type="iam.projects.service_accounts.key",
        method_name="get",
        method_params={"ctx": ctx, "name": resource_id},
    )

    ret = await hub.tool.gcp.generate.generic_exec.execute(execution_context)

    result["comment"] += ret["comment"]
    if not ret["result"]:
        result["result"] = False
        return result

    result["ret"] = ret["ret"]
    return result


async def upload(hub, ctx, service_account_id: str, public_key_data: str):
    r"""Uploads the public key portion of a key pair that you manage, and associates the public key with a service account key.

    After you upload the public key, you can use the private key from the key pair as a service account key.

    Args:
        service_account_id(str):
            Resource id of the service account

        public_key_data(str):
            The public key to associate with the service account. Must be an RSA public key that is wrapped in an X.509 v3 certificate. Include the first line, -----BEGIN CERTIFICATE-----, and the last line, -----END CERTIFICATE-----. A base64-encoded string.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.iam.projects.service_accounts.key.upload
              - kwargs:
                  service_account_id: projects/{...}/serviceAccounts/{...}
                  public_key_data: <base64-encoded X.509 v3 certificate>
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    if not service_account_id:
        result["result"] = False
        result["comment"] = [
            f"{RESOURCE_TYPE_FULL}#upload(): service_account_id is required"
        ]
        return result

    if not public_key_data:
        result["result"] = False
        result["comment"] = [
            f"{RESOURCE_TYPE_FULL}#upload(): public_key_data is required"
        ]
        return result

    execution_context = ExecutionContext(
        resource_type=RESOURCE_TYPE,
        method_name="upload",
        method_params={
            "ctx": ctx,
            "name": service_account_id,
            "body": {"public_key_data": public_key_data},
        },
    )

    ret = await hub.tool.gcp.generate.generic_exec.execute(execution_context)

    result["comment"] += ret["comment"]
    if not ret["result"]:
        result["result"] = False
        return result

    result["ret"] = ret["ret"]
    return result


async def disable(hub, ctx, resource_id: str):
    r"""Disable a ServiceAccountKey. A disabled service account key can be re-enabled.

    Args:
        resource_id(str):
            Resource id of the service account key

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.iam.projects.service_accounts.key.disable
              - kwargs:
                  resource_id: projects/{...}/serviceAccounts/{...}/keys/{...}
    """

    def validate(get_ret):
        assert get_ret.get("disabled")

    return await _api_op(hub, ctx, "disable", resource_id, validate)


async def enable(hub, ctx, resource_id: str):
    r"""Enable a ServiceAccountKey.

    Args:
        resource_id(str):
            Resource id of the service account key

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.iam.projects.service_accounts.key.enable
              - kwargs:
                  resource_id: projects/{...}/serviceAccounts/{...}/keys/{...}
    """

    def validate(get_ret):
        assert "disabled" not in get_ret

    return await _api_op(hub, ctx, "enable", resource_id, validate)


async def _api_op(hub, ctx, api_method: str, resource_id: str, validate: Callable):
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    if not resource_id:
        result["result"] = False
        result["comment"] = [
            f"{RESOURCE_TYPE_FULL}#{api_method}(): resource_id is required"
        ]
        return result

    async def validate_resource():
        """Validate subsequent get requests return expected output."""
        get_ret = await hub.tool.gcp.generate.generic_exec.execute(
            ExecutionContext(
                resource_type=RESOURCE_TYPE,
                method_name="get",
                method_params={"ctx": ctx, "name": resource_id},
            )
        )
        try:
            assert get_ret["result"]
            validate(get_ret["ret"])
        except AssertionError:
            # return dummy rerun_data to activate reconciliation
            result["rerun_data"] = {"api_method": api_method}
            result["comment"] += get_ret["comment"]
            result["result"] = False
            return result

        result["ret"] = get_ret["ret"]
        return result

    if ctx.get("rerun_data"):
        return await validate_resource()

    execution_context = ExecutionContext(
        resource_type=RESOURCE_TYPE,
        method_name=api_method,
        method_params={"ctx": ctx, "name": resource_id},
    )

    ret = await hub.tool.gcp.generate.generic_exec.execute(execution_context)

    result["comment"] += ret["comment"]
    if not ret["result"]:
        result["result"] = False
        return result

    return await validate_resource()



"""Exec module for managing Firewalls."""
__func_alias__ = {"list_": "list"}

from typing import Any
from typing import Dict


async def list_(
    hub,
    ctx,
    max_results: int = None,
    project: str = None,
    filter_: (str, "alias=filter") = None,
    order_by: str = None,
    page_token: str = None,
    return_partial_success: bool = None,
) -> Dict[str, Any]:
    r"""Retrieves the list of firewall rules available to the specified project.

    Args:
        max_results(int, Optional):
            The maximum number of results per page that should be returned. If the number of available results is larger
            than `maxResults`, Compute Engine returns a `nextPageToken` that can be used to get the next page of results
            in subsequent list requests. Acceptable values are `0` to `500`, inclusive. (Default: `500`). Defaults to 500.
        order_by(str, Optional):
            Sorts list results by a certain order. By default, results are returned in alphanumerical order based on the
            resource name. You can also sort results in descending order based on the creation timestamp using
            `orderBy="creationTimestamp desc"`. This sorts results based on the `creationTimestamp` field in reverse
            chronological order (newest result first). Use this to sort resources like operations so that the newest
            operation is returned first. Currently, only sorting by `name` or `creationTimestamp desc` is supported.
            Defaults to None.
        page_token(str, Optional):
            Specifies a page token to use. Set `pageToken` to the `nextPageToken` returned by a previous list request to
            get the next page of results. Defaults to None.
        return_partial_success(bool, Optional):
            Opt-in for partial success behavior which provides partial results in case of failure. The default value is
            false. Defaults to None.
        filter(str, Optional):
            A filter expression that filters resources listed in the response. Most Compute resources support two types
            of filter expressions: expressions that support regular expressions and expressions that follow API
            improvement proposal AIP-160. If you want to use AIP-160, your expression must specify the field name, an
            operator, and the value that you want to use for filtering. The value must be a string, a number, or a
            boolean. The operator must be either `=`, `!=`, `>`, `<`, `<=`, `>=` or `:`. For example, if you are
            filtering Compute Engine instances, you can exclude instances named `example-instance` by specifying
            `name != example-instance`. The `:` operator can be used with string fields to match substrings. For
            non-string fields it is equivalent to the `=` operator. The `:*` comparison can be used to test whether a
            key has been defined. For example, to find all objects with `owner` label use: ``` labels.owner:* ``` You
            can also filter nested fields. For example, you could specify `scheduling.automaticRestart = false` to
            include instances only if they are not scheduled for automatic restarts. You can use filtering on nested
            fields to filter based on resource labels. To filter on multiple expressions, provide each separate
            expression within parentheses. For example: ``` (scheduling.automaticRestart = true) (cpuPlatform = "Intel Skylake") ```
            By default, each expression is an `AND` expression. However, you can include `AND` and `OR` expressions
            explicitly. For example: ``` (cpuPlatform = "Intel Skylake") OR (cpuPlatform = "Intel Broadwell") AND (scheduling.automaticRestart = true) ```
            If you want to use a regular expression, use the `eq` (equal) or `ne` (not equal) operator against a single
            un-parenthesized expression with or without quotes or against multiple parenthesized expressions.

            Examples: `fieldname eq unquoted literal` `fieldname eq 'single quoted literal'`
            `fieldname eq "double quoted literal"` `(fieldname1 eq literal) (fieldname2 ne "literal")`
            The literal value is interpreted as a regular expression using Google RE2 library syntax. The literal value
            must match the entire field. For example, to filter for instances that do not end with name "instance", you
            would use `name ne .*instance`. Defaults to None.
        project(str):
            Project ID for this request.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

        list-firewalls:
            exec.run:
               - path: gcp.compute.firewall.list
               - kwargs:
                   project: project-name
    """
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    ret = await hub.exec.gcp_api.client.compute.firewall.list(
        ctx,
        project=project,
        max_results=max_results,
        filter=filter_,
        orderBy=order_by,
        page_token=page_token,
        return_partial_success=return_partial_success,
    )
    if not ret["result"]:
        result["comment"] += ret["comment"]
        result["result"] = False
        return result

    result["ret"] = ret["ret"].get("items", [])
    return result


async def get(
    hub,
    ctx,
    project: str = None,
    name: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Returns the specified firewall.

    Args:
        project: Project ID for this request.
        name: Name of the firewall rule to return.
        resource_id: Idem resource ID.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

        get-firewall:
            exec.run:
               - path: gcp.compute.firewall.get
               - kwargs:
                   project: project-name
                   name: firewall-name
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if resource_id:
        ret = await hub.exec.gcp_api.client.compute.firewall.get(
            ctx,
            resource_id=resource_id,
        )
    elif project and name:
        ret = await hub.exec.gcp_api.client.compute.firewall.get(
            ctx, project=project, firewall=name
        )
    else:
        result["result"] = False
        result["comment"] = [
            f"gcp.compute.firewall#get(): either resource_id or project and firewall"
            f" should be specified."
        ]
        return result

    result["comment"] += ret["comment"]
    if not ret["result"]:
        result["result"] = False
        return result

    result["ret"] = ret["ret"]
    return result



"""Metadata module for managing Buckets."""
# this will be autogenerated by pop-create with the resource id based on the path value from JSON schema for GET
PATH = "b/{bucket}"

NATIVE_RESOURCE_TYPE = "storage.buckets"



"""State module for managing Snapshots."""
__contracts__ = ["resource"]

from dataclasses import make_dataclass, field
from typing import Dict, Any, List

from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from idem_gcp.tool.gcp.utils import global_absent

# prevent commit hook from removing the import
absent = global_absent

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    project: str = None,
    resource_id: str = None,
    source_disk: str = None,
    storage_locations: List[str] = None,
    location_hint: str = None,
    label_fingerprint: str = None,
    description: str = None,
    labels: Dict[str, Any] = None,
    source_disk_encryption_key: make_dataclass(
        "CustomerEncryptionKey",
        [
            ("kms_key_service_account", str, field(default=None)),
            ("rsa_encrypted_key", str, field(default=None)),
            ("kms_key_name", str, field(default=None)),
            ("raw_key", str, field(default=None)),
        ],
    ) = None,
    chain_name: str = None,
    snapshot_type: str = None,
    snapshot_encryption_key: make_dataclass(
        "CustomerEncryptionKey",
        [
            ("kms_key_service_account", str, field(default=None)),
            ("rsa_encrypted_key", str, field(default=None)),
            ("kms_key_name", str, field(default=None)),
            ("raw_key", str, field(default=None)),
        ],
    ) = None,
    request_id: str = None,
) -> Dict[str, Any]:
    r"""Creates a snapshot in the specified project using the data included in the request.

    For regular snapshot creation, consider using this method instead of disks.createSnapshot, as this method supports more features, such as creating snapshots in a project different from the source disk project.

    Args:
        storage_locations(List[str], Optional):
            Cloud Storage bucket storage location of the snapshot (regional or multi-regional). Defaults to None.

        location_hint(str, Optional):
            An opaque location hint used to place the snapshot close to other resources. This field is for use by internal tools that use the public API. Defaults to None.

        label_fingerprint(str, Optional):
            A fingerprint for the labels being applied to this snapshot, which is essentially a hash of the labels set used for optimistic locking. The fingerprint is initially generated by Compute Engine and changes after every request to modify or update labels. You must always provide an up-to-date fingerprint hash in order to update or change labels, otherwise the request will fail with error 412 conditionNotMet. To see the latest fingerprint, make a get() request to retrieve a snapshot. Defaults to None.

        description(str, Optional):
            An optional description of this resource. Provide this property when you create the resource. Defaults to None.

        labels(Dict[str, Any], Optional):
            Labels to apply to this snapshot. These can be later modified by the setLabels method. Label values may be empty. Defaults to None.

        source_disk_encryption_key(Dict[str, Any], Optional):
            The customer-supplied encryption key of the source disk. Required if the source disk is protected by a customer-supplied encryption key. Defaults to None.
            * kms_key_service_account (str, Optional): The service account being used for the encryption request for the given KMS key. If absent, the Compute Engine default service account is used. For example: "kmsKeyServiceAccount": "name@project_id.iam.gserviceaccount.com/
            * sha256 (str, Optional): [Output only] The RFC 4648 base64 encoded SHA-256 hash of the customer-supplied encryption key that protects this resource.
            * rsa_encrypted_key (str, Optional): Specifies an RFC 4648 base64 encoded, RSA-wrapped 2048-bit customer-supplied encryption key to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rsaEncryptedKey": "ieCx/NcW06PcT7Ep1X6LUTc/hLvUDYyzSZPPVCVPTVEohpeHASqC8uw5TzyO9U+Fka9JFH z0mBibXUInrC/jEk014kCK/NPjYgEMOyssZ4ZINPKxlUh2zn1bV+MCaTICrdmuSBTWlUUiFoD D6PYznLwh8ZNdaheCeZ8ewEXgFQ8V+sDroLaN3Xs3MDTXQEMMoNUXMCZEIpg9Vtp9x2oe==" The key must meet the following requirements before you can provide it to Compute Engine: 1. The key is wrapped using a RSA public key certificate provided by Google. 2. After being wrapped, the key must be encoded in RFC 4648 base64 encoding. Gets the RSA public key certificate provided by Google at: https://cloud-certs.storage.googleapis.com/google-cloud-csek-ingress.pem
            * kms_key_name (str, Optional): The name of the encryption key that is stored in Google Cloud KMS. For example: "kmsKeyName": "projects/kms_project_id/locations/region/keyRings/ key_region/cryptoKeys/key
            * raw_key (str, Optional): Specifies a 256-bit customer-supplied encryption key, encoded in RFC 4648 base64 to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rawKey": "SGVsbG8gZnJvbSBHb29nbGUgQ2xvdWQgUGxhdGZvcm0="

        source_disk(str, Optional):
            The source disk used to create this snapshot. Defaults to None.

        chain_name(str, Optional):
            Creates the new snapshot in the snapshot chain labeled with the specified name. The chain name must be 1-63 characters long and comply with RFC1035. This is an uncommon option only for advanced service owners who needs to create separate snapshot chains, for example, for chargeback tracking. When you describe your snapshot resource, this field is visible only if it has a non-empty value. Defaults to None.

        snapshot_type(str, Optional):
            Indicates the type of the snapshot.
            Enum type. Allowed values:
            "ARCHIVE"
            "STANDARD". Defaults to None.

        snapshot_encryption_key(Dict[str, Any], Optional):
            Encrypts the snapshot using a customer-supplied encryption key. After you encrypt a snapshot using a customer-supplied key, you must provide the same key if you use the snapshot later. For example, you must provide the encryption key when you create a disk from the encrypted snapshot in a future request. Customer-supplied encryption keys do not protect access to metadata of the snapshot. If you do not provide an encryption key when creating the snapshot, then the snapshot will be encrypted using an automatically generated key and you do not need to provide a key to use the snapshot later. Defaults to None.
            * kms_key_service_account (str, Optional): The service account being used for the encryption request for the given KMS key. If absent, the Compute Engine default service account is used. For example: "kmsKeyServiceAccount": "name@project_id.iam.gserviceaccount.com/
            * sha256 (str, Optional): [Output only] The RFC 4648 base64 encoded SHA-256 hash of the customer-supplied encryption key that protects this resource.
            * rsa_encrypted_key (str, Optional): Specifies an RFC 4648 base64 encoded, RSA-wrapped 2048-bit customer-supplied encryption key to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rsaEncryptedKey": "ieCx/NcW06PcT7Ep1X6LUTc/hLvUDYyzSZPPVCVPTVEohpeHASqC8uw5TzyO9U+Fka9JFH z0mBibXUInrC/jEk014kCK/NPjYgEMOyssZ4ZINPKxlUh2zn1bV+MCaTICrdmuSBTWlUUiFoD D6PYznLwh8ZNdaheCeZ8ewEXgFQ8V+sDroLaN3Xs3MDTXQEMMoNUXMCZEIpg9Vtp9x2oe==" The key must meet the following requirements before you can provide it to Compute Engine: 1. The key is wrapped using a RSA public key certificate provided by Google. 2. After being wrapped, the key must be encoded in RFC 4648 base64 encoding. Gets the RSA public key certificate provided by Google at: https://cloud-certs.storage.googleapis.com/google-cloud-csek-ingress.pem
            * kms_key_name (str, Optional): The name of the encryption key that is stored in Google Cloud KMS. For example: "kmsKeyName": "projects/kms_project_id/locations/region/keyRings/ key_region/cryptoKeys/key
            * raw_key (str, Optional): Specifies a 256-bit customer-supplied encryption key, encoded in RFC 4648 base64 to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rawKey": "SGVsbG8gZnJvbSBHb29nbGUgQ2xvdWQgUGxhdGZvcm0="

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.

        project(str):
            Project ID for this request.

        name(str):
            Name of the Snapshot resource to return.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              gcp.compute.snapshot.present:
                - name: value
                - project: value
                - snapshot: value
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if ctx.get("wrapper_result"):
        result = ctx.get("wrapper_result")

    # to be autogenerated by pop-create based on insert/update props in properties.yaml
    resource_body = {
        "name": name,
        "storage_locations": storage_locations,
        "location_hint": location_hint,
        "label_fingerprint": label_fingerprint,
        "description": description,
        "labels": labels,
        "source_disk_encryption_key": source_disk_encryption_key,
        "source_disk": source_disk,
        "chain_name": chain_name,
        "snapshot_type": snapshot_type,
        "snapshot_encryption_key": snapshot_encryption_key,
    }

    resource_body = {k: v for (k, v) in resource_body.items() if v is not None}
    operation = None

    if result["old_state"]:
        resource_body["label_fingerprint"] = label_fingerprint or result[
            "old_state"
        ].get("label_fingerprint")

        changes = hub.tool.gcp.utils.compare_states(
            result["old_state"],
            {"resource_id": resource_id, **resource_body},
            "compute.snapshot",
        )

        # This check is necessary because "changes" does not detect label removal
        are_labels_changed = (
            labels is not None
            and not (labels == {} and result["old_state"].get("labels") is None)
            and result["old_state"].get("labels") != labels
        )

        if not changes and not are_labels_changed:
            result["comment"].append(
                hub.tool.gcp.comment_utils.up_to_date_comment(
                    "gcp.compute.snapshot", name
                )
            )
            result["new_state"] = result["old_state"]
            return result

        changed_non_updatable_properties = (
            hub.tool.gcp.resource_prop_utils.get_changed_non_updatable_properties(
                "compute.snapshot", changes
            )
        )

        if changed_non_updatable_properties:
            result["result"] = False
            result["comment"].append(
                hub.tool.gcp.comment_utils.non_updatable_properties_comment(
                    "gcp.compute.snapshot",
                    name,
                    changed_non_updatable_properties,
                )
            )
            result["new_state"] = result["old_state"]
            return result

        if ctx.get("test", False):
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_update_comment(
                    "gcp.compute.snapshot", name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                resource_body
            )
            result["new_state"]["resource_id"] = resource_id
            return result

        set_labels_body = {
            "labels": labels,
        }
        if label_fingerprint:
            set_labels_body.update(
                {
                    "label_fingerprint": label_fingerprint,
                }
            )
        elif result["old_state"].get("label_fingerprint"):
            set_labels_body.update(
                {
                    "label_fingerprint": result["old_state"].get("label_fingerprint"),
                }
            )

        # the setLabels() method accepts "resource" as the name of the snapshot, so
        # the resource_id is split into its properties ("project" and "snapshot") here to map "resource" to "snapshot
        resource_id_props = (
            hub.tool.gcp.resource_prop_utils.get_elements_from_resource_id(
                "compute.snapshot", resource_id
            )
        )
        update_ret = await hub.exec.gcp_api.client.compute.snapshot.setLabels(
            ctx,
            project=resource_id_props.get("project"),
            resource=resource_id_props.get("snapshot"),
            body=set_labels_body,
        )

        if not update_ret["result"] or not update_ret["ret"]:
            result["result"] = False
            result["comment"] += update_ret["comment"]
            return result

        if hub.tool.gcp.operation_utils.is_operation(update_ret["ret"]):
            operation = update_ret["ret"]

    else:
        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    "gcp.compute.snapshot", name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                resource_body
            )
            result["new_state"]["resource_id"] = resource_id
            return result

        # Create
        create_ret = await hub.exec.gcp_api.client.compute.snapshot.insert(
            ctx,
            name=name,
            project=project,
            request_id=request_id,
            body=resource_body,
        )
        if not create_ret["result"]:
            result["result"] = False
            if create_ret["comment"] and next(
                (
                    comment
                    for comment in create_ret["comment"]
                    if "alreadyExists" in comment
                ),
                None,
            ):
                result["comment"].append(
                    hub.tool.gcp.comment_utils.already_exists_comment(
                        "gcp.compute.snapshot", name
                    )
                )
                return result
            else:
                result["comment"] += create_ret["comment"]

        if create_ret["ret"]:
            if hub.tool.gcp.operation_utils.is_operation(create_ret["ret"]):
                operation = create_ret["ret"]

    if operation:
        operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
            operation.get("selfLink"), "compute.global_operation"
        )
        result["rerun_data"] = {
            "operation_id": operation_id,
            "old_state": result["old_state"],
        }

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves the list of Snapshot resources contained within the specified project.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.snapshot
    """
    result = {}

    describe_ret = await hub.exec.gcp.compute.snapshot.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(f"Could not describe snapshots {describe_ret['comment']}")
        return {}

    for resource in describe_ret["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "gcp.compute.snapshot.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



"""Exec module for managing Cloud Key Management Service crypto keys."""
from typing import Any
from typing import Dict


__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    resource_id: str,
):
    """Returns a crypto key by its Idem resource ID.

    Args:
        resource_id(str):
            Idem resource ID. ``projects/{project id}/locations/{location id}/keyRings/{keyRing}/cryptoKeys/{cryptoKey}``

    Returns:
        CryptoKey resource

    Examples:
        .. code-block:: sls

            {% set project_id = 'project-name' %}
            {% set location_id = 'us-east1' %}
            {% set key_ring = 'key-ring' %}
            {% set crypto_key = 'crypto-key' %}
            get-crypto-key:
                exec.run:
                    - path: gcp.cloudkms.crypto_key.get
                    - kwargs:
                        resource_id: projects/{{project_id}}/locations/{{location_id}}/keyRings/{{key_ring}}/cryptoKeys/{{crypto_key}}
    """
    result = {
        "comment": [],
        "ret": [],
        "result": True,
    }

    crypto_key = await hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings.crypto_keys.get(
        ctx, _name=resource_id
    )

    if not crypto_key["result"]:
        result["comment"] += crypto_key["comment"]
        result["result"] = False
        return result

    result["ret"] = crypto_key["ret"]

    if not result["ret"]:
        result["comment"] += (
            hub.tool.gcp.comment_utils.get_empty_comment(
                "gcp.cloudkms.crypto_key", resource_id
            ),
        )

    return result


async def list_(
    hub, ctx, key_ring: str, filter_: (str, "alias=filter") = None, order_by: str = None
) -> Dict[str, Any]:
    r"""Retrieves the crypto keys in a key ring.

    Args:
        key_ring(str):
            key ring resource_id.

        filter(str, Optional):
            Only include resources that match the filter in the response. For more information, see
            `Sorting and filtering list results`_.

        order_by(str, Optional):
            Specify how the results should be sorted. If not specified, the results will be sorted in the default order.
            For more information, see `Sorting and filtering list results`_.

    .. _Sorting and filtering list results: https://cloud.google.com/kms/docs/sorting-and-filtering

    Examples:
        .. code-block:: sls

            list-locations:
                exec.run:
                   - path: gcp.cloudkms.location.list
                   - kwargs:
                         project: project-name

            #!require:list-locations
            list-key-rings:
                exec.run:
                   - path: gcp.cloudkms.key_ring.list
                   - kwargs:
                         location: {% for v in hub.idem.arg_bind.resolve('${exec:list-locations}') -%}
                                       {{ v['resource_id'] if v.get('display_name') == 'South Carolina' }}
                                   {%- endfor %}

            #!require:list-key-rings
            list-crypto-keys:
                exec.run:
                   - path: gcp.cloudkms.crypto_key.list
                   - kwargs:
                         key_ring: ${exec:list-key-rings:[0]:resource_id}

            #!END

            list-crypto-keys-filtered:
                exec.run:
                   - path: gcp.cloudkms.crypto_key.list
                   - kwargs:
                         key_ring: projects/project-name/locations/global/keyRings/kr-global-test
                         filter: nextRotationTime < 2023-10-02
    """
    result = {
        "comment": [],
        "ret": [],
        "result": True,
    }

    crypto_keys = await hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings.crypto_keys.list(
        ctx, parent=key_ring, filter=filter_, orderBy=order_by
    )
    if not crypto_keys["result"]:
        result["comment"] += crypto_keys["comment"]
        result["result"] = False
        return result

    result["ret"] = crypto_keys["ret"].get("items", [])

    return result



"""State module for managing Reservations."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from idem_gcp.tool.gcp.utils import zonal_absent

# prevent commit hook from removing the import
absent = zonal_absent

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    request_id: str = None,
    project: str = None,
    zone: str = None,
    description: str = None,
    share_settings: make_dataclass(
        "ShareSettings",
        [
            (
                "project_map",
                make_dataclass(
                    "ShareSettingsProjectConfig",
                    [("project_id", str, field(default=None))],
                ),
                field(default=None),
            ),
            ("share_type", str, field(default="LOCAL")),
        ],
    ) = None,
    specific_reservation: make_dataclass(
        "AllocationSpecificSKUReservation",
        [
            ("count", str, field(default=None)),
            (
                "instance_properties",
                make_dataclass(
                    "AllocationSpecificSKUAllocationReservedInstanceProperties",
                    [
                        ("machine_type", str, field(default=None)),
                        (
                            "local_ssds",
                            List[
                                make_dataclass(
                                    "AllocationSpecificSKUAllocationAllocatedInstancePropertiesReservedDisk",
                                    [
                                        ("disk_size_gb", str, field(default=None)),
                                        ("interface", str, field(default=None)),
                                    ],
                                )
                            ],
                            field(default=None),
                        ),
                        ("location_hint", str, field(default=None)),
                        ("guest_accelerators", str, field(default=None)),
                        ("min_cpu_platform", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    specific_reservation_required: bool = False,
) -> Dict[str, Any]:
    r"""Creates a new reservation. For more information, read Reserving zonal resources.

    Args:
        name(str):
            An Idem name of the resource.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.

        project(str, Optional):
            Project ID for this request.

        zone(str, Optional):
            Name of the zone for this request.

        description(str, Optional):
            An optional description of this resource. Provide this field when you create the resource. Defaults to None.

        share_settings(ShareSettings, Optional):
            Share-settings for shared-reservation

        specific_reservation(AllocationSpecificSKUReservation, Optional):
            Reservation for instances with specific machine shapes.

        specific_reservation_required(bool, Optional):
            Indicates whether the reservation can be consumed by VMs with affinity for \"any\" reservation. If the field is set, then only VMs that target the reservation by name can consume from this reservation.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    # TODO uncomment below line, when implementation is added
    # project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    result["comment"].append(
        "No-op: There is no create/update function for gcp.compute.reservation"
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves a list of all the reservations that have been configured for the specified project in specified zone.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.reservation
    """
    result = {}

    # TODO: Pagination
    describe_ret = await hub.exec.gcp_api.client.compute.reservation.aggregatedList(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(
            f"Could not describe gcp.compute.reservation {describe_ret['comment']}"
        )
        return {}

    for resource in describe_ret["ret"].get("items", []):
        resource_id = resource.get("resource_id")

        result[resource_id] = {
            "gcp.compute.reservation.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



# VMware Idem Plugin
# Copyright (c) 2020-2022 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any
from typing import Dict

try:
    from google.oauth2.service_account import Credentials as Oauth2Credentials

    import google.auth._service_account_info as service_account_info

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def gather(hub, profiles) -> Dict[str, Any]:
    """
    Get profile names from encrypted service account data

    Example:
    .. code-block:: yaml

        gcp:
          service_account:
            default:
              type: "service_account"
              project_id: "idem_gcp.com:idemgcp"
              private_key_id: "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
              private_key: "-----BEGIN PRIVATE KEY-----\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxn\n-----END PRIVATE KEY-----\n"
              client_email: "idem-gcp@example.com.iam.gserviceaccount.com"
              client_id: "xxxxxxxxxxxxxxxxxxxxx"
              auth_uri: "https://accounts.google.com/o/oauth2/auth"
              token_uri: "https://oauth2.googleapis.com/token"
              auth_provider_x509_cert_url: "https://www.googleapis.com/oauth2/v1/certs"
              client_x509_cert_url: "https://www.googleapis.com/robot/v1/metadata/x509/idem-gcp%40example.com.iam.gserviceaccount.com"
            my-project:
              type: "searvice_account"
              project_id: "idem_gcp.com:my-project"
              private_key_id: "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
              private_key: "-----BEGIN PRIVATE KEY-----\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxn\n-----END PRIVATE KEY-----\n"
              client_email: "idem-gcp@example.com.iam.gserviceaccount.com"
              client_id: "xxxxxxxxxxxxxxxxxxxxx"
              auth_uri: "https://accounts.google.com/o/oauth2/auth"
              token_uri: "https://oauth2.googleapis.com/token"
              auth_provider_x509_cert_url: "https://www.googleapis.com/oauth2/v1/certs"
              client_x509_cert_url: "https://www.googleapis.com/robot/v1/metadata/x509/idem-gcp%40example.com.iam.gserviceaccount.com"
    """
    sub_profiles = {}
    credentials = {}
    for profile, creds in profiles.get("gcp", {}).items():
        signer = service_account_info.from_dict(
            creds, require=["client_email", "token_uri"]
        )
        credentials = Oauth2Credentials(
            signer,
            service_account_email=creds["client_email"],
            token_uri=creds["token_uri"],
            project_id=creds["project_id"],
        )
        sub_profiles[profile] = {"credentials": credentials}

    return sub_profiles



import json
import os
import shutil
import subprocess
import tempfile

script_template = """
cd "{clone_dir}"
git clone "{official_repo_url}"
cd "{work_dir}"
git checkout --quiet {rev}
python -m venv venv
source "{work_dir}"/venv/bin/activate
pip uninstall -q -y --disable-pip-version-check --require-virtualenv idem-gcp
pip install -q --disable-pip-version-check --require-virtualenv -e .
idem doc states.gcp --output=json > "{out_file}"
"""


def get_idem_doc(idem_cli, doc_path):
    return idem_cli("doc", doc_path).json


def fetch_current_states_schema(hub, idem_cli):
    return get_idem_doc(idem_cli, "states.gcp")


def fetch_current_exec_schema(hub, idem_cli):
    return get_idem_doc(idem_cli, "exec.gcp")


# !!! important - currently you can't fetch schemas from other than the latest changeset in forked repos
def fetch_states_schema_from_official_repo(hub, official_repo_url: str, rev: str):
    tmp_dir = tempfile.mkdtemp()
    try:
        return json.loads(_fetch_schema_for_rev(official_repo_url, rev, tmp_dir))
    finally:
        shutil.rmtree(tmp_dir, ignore_errors=True)


def _fetch_schema_for_rev(official_repo_url: str, rev: str, clone_dir: str):
    out_file = os.path.abspath(os.path.join(clone_dir, f"{rev}.json"))
    work_dir = os.path.join(
        clone_dir, os.path.splitext(os.path.basename(official_repo_url))[0]
    )

    script_file = f"schema-{rev}.sh"
    command = ["/bin/bash", script_file]

    with open(os.path.join(clone_dir, script_file), "w") as f:
        script = script_template.format(
            **{
                "clone_dir": clone_dir,
                "official_repo_url": official_repo_url,
                "rev": rev,
                "out_file": out_file,
                "work_dir": work_dir,
            }
        )
        f.write(script)

    print(f"Fetching schema for commit/tag {rev}")

    ret = subprocess.run(
        command,
        cwd=clone_dir,
        capture_output=True,
    )

    if not ret or ret.returncode != 0:
        print(f"Error fetching schema: {str(ret.stderr, 'utf-8')}")
        return None

    with open(os.path.join(work_dir, out_file)) as f:
        result = f.read()

    if result:
        end_mark = result.rfind("}")
        if end_mark != -1:
            result = result[: end_mark + 1]

    return result



# VMware Idem Plugin
# Copyright (c) 2020-2022 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from os import environ
from typing import Any
from typing import Dict

try:
    from google.oauth2.service_account import Credentials as Oauth2Credentials

    import google.auth._service_account_info as service_account_info

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def gather(hub) -> Dict[str, Any]:
    """
    Sets the default profile from the environment. All information can
    be obtained from a standard GCP credentials JSON file. The variables
    required are:

        GCP_AUTH_CLIENT_EMAIL: The service account client e-mail address.
        GCP_AUTH_PRIVATE_KEY_ID: The service account private key id.
        GCP_AUTH_PRIVATE_KEY: The service account private key.
        GCP_AUTH_PROJECT_ID: The id of the project in which the service account.
        GCP_AUTH_TOKEN_URI: The service account token URI.

    Note: The credential type is always "service_account".
    """
    sub_profiles = {}
    credentials = {}
    profile_name = "default"

    creds = {}
    try:
        creds["type"] = "service_account"
        creds["project_id"] = environ["GCP_AUTH_PROJECT_ID"]
        creds["private_key_id"] = environ["GCP_AUTH_PRIVATE_KEY_ID"]
        creds["client_email"] = environ["GCP_AUTH_CLIENT_EMAIL"]
        creds["private_key"] = environ["GCP_AUTH_PRIVATE_KEY"]
        creds["token_uri"] = environ["GCP_AUTH_TOKEN_URI"]
    except KeyError:
        return {}

    signer = service_account_info.from_dict(
        creds, require=["client_email", "token_uri"]
    )
    credentials = Oauth2Credentials(
        signer,
        service_account_email=creds["client_email"],
        token_uri=creds["token_uri"],
        project_id=creds["project_id"],
    )
    sub_profiles[profile_name] = {"credentials": credentials}

    return sub_profiles



import aiohttp


async def gather(hub):
    """
    Authenticate with a username and password to the given endpoint url.
    Any extra parameters will be saved as part of the profile.

    Example:

    .. code-block:: sls

        gcp.basic_auth:
          profile_name:
            username: my_user
            password: my_token
            endpoint_url: https://gcp.com
    """
    sub_profiles = {}
    for profile, ctx in hub.acct.PROFILES.get("gcp.basic_auth", {}).items():
        sub_profiles[profile] = dict(
            endpoint_url=ctx.pop("endpoint_url"),
            auth=aiohttp.BasicAuth(ctx.pop("username"), ctx.pop("password")),
            **ctx
        )

    return sub_profiles



"""Exec module for managing Instance Groups."""

__func_alias__ = {"list_": "list"}

from dataclasses import field
from dataclasses import make_dataclass
from typing import List

from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext
from idem_gcp.tool.gcp.generate.exec_param import ExecParam
from idem_gcp.tool.gcp.generate.scope import Scope


async def list_(
    hub,
    ctx,
    project: str = None,
    zone: str = None,
    filter_: (str, "alias=filter") = None,
    order_by: str = None,
):
    r"""Retrieves the list of zonal instance group resources contained within the specified zone. For managed instance groups, use the instanceGroupManagers or regionInstanceGroupManagers methods instead.

    Args:
        project(str, Optional):
            Project ID for this request.

        zone(str, Optional):
            The name of the zone where the instance group is located.

        filter(str, Optional):
            A filter expression that filters resources listed in the response. Most Compute resources support two types of filter expressions: expressions that support regular expressions and expressions that follow API improvement proposal AIP-160. If you want to use AIP-160, your expression must specify the field name, an operator, and the value that you want to use for filtering. The value must be a string, a number, or a boolean. The operator must be either `=`, `!=`, `>`, `<`, `<=`, `>=` or `:`. For example, if you are filtering Compute Engine instances, you can exclude instances named `example-instance` by specifying `name != example-instance`. The `:` operator can be used with string fields to match substrings. For non-string fields it is equivalent to the `=` operator. The `:*` comparison can be used to test whether a key has been defined. For example, to find all objects with `owner` label use: ``` labels.owner:* ``` You can also filter nested fields. For example, you could specify `scheduling.automaticRestart = false` to include instances only if they are not scheduled for automatic restarts. You can use filtering on nested fields to filter based on resource labels. To filter on multiple expressions, provide each separate expression within parentheses. For example: ``` (scheduling.automaticRestart = true) (cpuPlatform = \"Intel Skylake\") ``` By default, each expression is an `AND` expression. However, you can include `AND` and `OR` expressions explicitly. For example: ``` (cpuPlatform = \"Intel Skylake\") OR (cpuPlatform = \"Intel Broadwell\") AND (scheduling.automaticRestart = true) ``` If you want to use a regular expression, use the `eq` (equal) or `ne` (not equal) operator against a single un-parenthesized expression with or without quotes or against multiple parenthesized expressions. Examples: `fieldname eq unquoted literal` `fieldname eq 'single quoted literal'` `fieldname eq \"double quoted literal\"` `(fieldname1 eq literal) (fieldname2 ne \"literal\")` The literal value is interpreted as a regular expression using Google RE2 library syntax. The literal value must match the entire field. For example, to filter for instances that do not end with name "instance", you would use `name ne .*instance`.

        order_by(str, Optional):
            Sorts list results by a certain order. By default, results are returned in alphanumerical order based on the resource name. You can also sort results in descending order based on the creation timestamp using `orderBy=\"creationTimestamp desc\"`. This sorts results based on the `creationTimestamp` field in reverse chronological order (newest result first). Use this to sort resources like operations so that the newest operation is returned first. Currently, only sorting by `name` or `creationTimestamp desc` is supported.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.instance_group.list
              - kwargs:
                  project: project-name
                  zone: zone-name
    """
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    # GCP method name is simply the name of this func or the one from __func_alias - we can pass it
    # also we can pass the resource name based on the file name
    execution_context = ExecutionContext(
        resource_type="compute.instance_group",
        method_name="list",
        method_params={
            "ctx": ctx,
            "project": project,
            "zone": zone,
            "filter": filter_,
            "order_by": order_by,
        },
        exec_params={
            ExecParam.SCOPED_FUNCTIONS: {
                Scope.ZONAL: "list",
                Scope.GLOBAL: "aggregatedList",
            }
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def get(
    hub,
    ctx,
    resource_id: str = None,
    project: str = None,
    zone: str = None,
    name: str = None,
):
    r"""Returns the specified zonal instance group. acceleratorTypes.get a list of available zonal instance groups by making a list() request. For managed instance groups, use the instanceGroupManagers or regionInstanceGroupManagers methods instead.

    Args:
        resource_id(str, Optional):
            An identifier of the resource in the provider.

        project(str, Optional):
            Project ID for this request.

        zone(str, Optional):
            Name of the zone for this request.

        name(str, Optional): The name of the instance group. Authorization requires the following IAM
        permission on the specified resource instanceGroup: compute.instance_group.get.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.instance_group.get
              - kwargs:
                  project: project-name
                  zone: zone-name
                  name: instance-name
    """
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    execution_context = ExecutionContext(
        resource_type="compute.instance_group",
        method_name="get",
        method_params={
            "ctx": ctx,
            "resource_id": resource_id,
            "project": project,
            "zone": zone,
            "instanceGroup": name,
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def list_instances(
    hub,
    ctx,
    instance_group_resource_id: str,
    filter_: (str, "alias=filter") = None,
    order_by: str = None,
):
    r"""Retrieves the list of instances in an instance groups.

    Args:
        instance_group_resource_id(str):
            Resource ID for the container instance group.

        filter(str, Optional):
            A filter expression that filters resources listed in the response. Most Compute resources support two types
            of filter expressions: expressions that support regular expressions and expressions that follow API
            improvement proposal AIP-160. If you want to use AIP-160, your expression must specify the field name, an
            operator, and the value that you want to use for filtering. The value must be a string, a number, or a
            boolean. The operator must be either `=`, `!=`, `>`, `<`, `<=`, `>=` or `:`. For example, if you are
            filtering Compute Engine instances, you can exclude instances named `example-instance` by specifying
            `name != example-instance`. The `:` operator can be used with string fields to match substrings. For
            non-string fields it is equivalent to the `=` operator. The `:*` comparison can be used to test whether a
            key has been defined. For example, to find all objects with `owner` label use: ``` labels.owner:* ``` You
            can also filter nested fields. For example, you could specify `scheduling.automaticRestart = false` to
            include instances only if they are not scheduled for automatic restarts. You can use filtering on nested
            fields to filter based on resource labels. To filter on multiple expressions, provide each separate
            expression within parentheses. For example: ``` (scheduling.automaticRestart = true) (cpuPlatform = \"Intel Skylake\") ```
            By default, each expression is an `AND` expression. However, you can include `AND` and `OR` expressions
            explicitly. For example: ``` (cpuPlatform = \"Intel Skylake\") OR (cpuPlatform = \"Intel Broadwell\") AND (scheduling.automaticRestart = true) ```
            If you want to use a regular expression, use the `eq` (equal) or `ne` (not equal) operator against a single
            un-parenthesized expression with or without quotes or against multiple parenthesized expressions.
            Examples: `fieldname eq unquoted literal` `fieldname eq 'single quoted literal'`
            `fieldname eq \"double quoted literal\"` `(fieldname1 eq literal) (fieldname2 ne \"literal\")` The literal
            value is interpreted as a regular expression using Google RE2 library syntax. The literal value must match
            the entire field. For example, to filter for instances that do not end with name "instance", you would use
            `name ne .*instance`.

        order_by(str, Optional):
            Sorts list results by a certain order. By default, results are returned in alphanumerical order based on the
            resource name. You can also sort results in descending order based on the creation timestamp using
            `orderBy=\"creationTimestamp desc\"`. This sorts results based on the `creationTimestamp` field in reverse
            chronological order (newest result first). Use this to sort resources like operations so that the newest
            operation is returned first. Currently, only sorting by `name` or `creationTimestamp desc` is supported.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.instance_group.list_instances
              - kwargs:
                    instance_group_resource_id: projects/project-name/zones/us-central1-a/instanceGroups/instance-group
    """
    # GCP method name is simply the name of this func or the one from __func_alias - we can pass it
    # also we can pass the resource name based on the file name
    els = hub.tool.gcp.resource_prop_utils.get_elements_from_resource_id(
        "compute.instance_group", instance_group_resource_id
    )
    execution_context = ExecutionContext(
        resource_type="compute.instance_group",
        method_name="listInstances",
        method_params={
            "ctx": ctx,
            "project": els["project"],
            "zone": els["zone"],
            "instance_group": els["instanceGroup"],
            "filter": filter_,
            "order_by": order_by,
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def remove_instances(
    hub,
    ctx,
    instance_group_resource_id: str,
    instances: List[str],
):
    r"""Removes instances from an instance group.

    Args:
        instance_group_resource_id(str):
            Resource ID for the container instance group.

        instances(List[str]):
            Instance IDs of the instances to be removed.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.instance_group.remove_instances
              - kwargs:
                    instance_group_resource_id: projects/project-name/zones/us-central1-a/instanceGroups/instance-group
                    instances:
                        - projects/project-name/zones/us-central1-a/instances/instance-1
    """
    # GCP method name is simply the name of this func or the one from __func_alias - we can pass it
    # also we can pass the resource name based on the file name
    els = hub.tool.gcp.resource_prop_utils.get_elements_from_resource_id(
        "compute.instance_group", instance_group_resource_id
    )
    execution_context = ExecutionContext(
        resource_type="compute.instance_group",
        method_name="removeInstances",
        method_params={
            "ctx": ctx,
            "project": els["project"],
            "zone": els["zone"],
            "instance_group": els["instanceGroup"],
            "body": {"instances": [{"instance": id} for id in instances]},
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def add_instances(
    hub,
    ctx,
    instance_group_resource_id: str,
    instances: List[str],
):
    r"""Adds instances to an instance group.

    Args:
        instance_group_resource_id(str):
            Resource ID for the container instance group.

        instances(List[str]):
            Instance IDs of the instances to be added.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.instance_group.add_instances
              - kwargs:
                    instance_group_resource_id: projects/project-name/zones/us-central1-a/instanceGroups/instance-group
                    instances:
                        - projects/project-name/zones/us-central1-a/instances/instance-1
    """
    # GCP method name is simply the name of this func or the one from __func_alias - we can pass it
    # also we can pass the resource name based on the file name
    els = hub.tool.gcp.resource_prop_utils.get_elements_from_resource_id(
        "compute.instance_group", instance_group_resource_id
    )
    execution_context = ExecutionContext(
        resource_type="compute.instance_group",
        method_name="addInstances",
        method_params={
            "ctx": ctx,
            "project": els["project"],
            "zone": els["zone"],
            "instance_group": els["instanceGroup"],
            "body": {"instances": [{"instance": id} for id in instances]},
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def set_named_ports(
    hub,
    ctx,
    instance_group_resource_id: str,
    named_ports: List[
        make_dataclass(
            "NamedPort",
            [("name", str, field(default=None)), ("port", int, field(default=None))],
        )
    ] = None,
    fingerprint: str = None,
):
    r"""Sets the named ports for the specified instance group.

    Args:
        instance_group_resource_id(str):
            Resource ID for the container instance group.

        named_ports(List[Dict[str, Any]]):
            The list of named ports to set for this instance group.
            Assigns a name to a port number. For example: {name: "http", port: 80} This allows the system to reference ports by the assigned name instead of a port number. Named ports can also contain multiple ports. For example: [{name: "app1", port: 8080}, {name: "app1", port: 8081}, {name: "app2", port: 8082}] Named ports apply to all instances in this instance group. . Defaults to None.
                * name (str, optional): The name for this named port. The name must be 1-63 characters long, and comply with RFC1035.
                * port (int, optional): The port number, which can be a value between 1 and 65535.

        fingerprint(str):
            The fingerprint of the named ports information for this instance group. Use this optional property to prevent conflicts when multiple users change the named ports settings concurrently. Obtain the fingerprint with the instanceGroups.get method. Then, include the fingerprint in your request to ensure that you do not overwrite changes that were applied from another concurrent request. A request with an incorrect fingerprint will fail with error 412 conditionNotMet.
            A base64-encoded string.

    Examples:
        .. code-block:: sls

            random-name:
              exec.run:
              - path: gcp.compute.instance_group.set_named_ports
              - kwargs:
                    instance_group_resource_id: projects/project-name/zones/us-central1-a/instanceGroups/instance-group
                    fingerprint: ABCD
                    named_ports:
                        - name: http
                          port: 8080
    """
    # GCP method name is simply the name of this func or the one from __func_alias - we can pass it
    # also we can pass the resource name based on the file name
    els = hub.tool.gcp.resource_prop_utils.get_elements_from_resource_id(
        "compute.instance_group", instance_group_resource_id
    )
    execution_context = ExecutionContext(
        resource_type="compute.instance_group",
        method_name="setNamedPorts",
        method_params={
            "ctx": ctx,
            "project": els["project"],
            "zone": els["zone"],
            "instance_group": els["instanceGroup"],
            "body": {
                "namedPorts": [np for np in named_ports],
                "fingerprint": fingerprint,
            },
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)



"""State module for managing Cloud Key Management Service locations."""
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
) -> Dict[str, Any]:
    """Present operation is not supported for this resource.

    Args:
        name:
            Idem name.

    Returns:
        .. code-block:: json

            {
                "result": False,
                "comment": "...",
            }

    """
    return {
        "result": False,
        "name": name,
        "old_state": None,
        "new_state": None,
        "comment": [
            hub.tool.gcp.comment_utils.no_resource_create_update_comment(
                "gcp.cloudkms.location"
            )
        ],
    }


async def absent(
    hub,
    ctx,
    name: str,
) -> Dict[str, Any]:
    """Absent operation is not supported for this resource.

    Args:
        name(str):
            Idem name.

    Returns:
        .. code-block:: json

            {
                "result": False,
                "comment": "...",
                "old_state": None,
                "new_state": None,
            }

    """
    return {
        "result": False,
        "name": name,
        "old_state": None,
        "new_state": None,
        "comment": [
            hub.tool.gcp.comment_utils.no_resource_create_update_comment(
                "gcp.cloudkms.location"
            )
        ],
    }


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Retrieve the list of available locations.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.cloudkms.location
    """
    result = {}

    locations = await hub.exec.gcp.cloudkms.location.list(
        ctx, project=ctx.acct.project_id
    )
    if not locations["result"]:
        hub.log.debug(
            f"Could not describe gcp.cloudkms.location in {ctx.acct.project_id} {locations['comment']}"
        )
        return {}

    for location in locations["ret"]:
        resource_id = location.get("resource_id")
        result[resource_id] = {
            "gcp.cloudkms.location.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in location.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



"""Google Cloud Platform firewalls state module.

Copyright (c) 2023 VMware, Inc. All Rights Reserved.
SPDX-License-Identifier: Apache-2.0

"""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from idem_gcp.tool.gcp.utils import global_absent

# prevent commit hook from removing the import
absent = global_absent

__contracts__ = ["resource"]
RESOURCE_TYPE = "compute.firewall"
RESOURCE_TYPE_FULL = "gcp.compute.firewall"


async def present(
    hub,
    ctx,
    name: str,
    project: str = None,
    target_tags: List[str] = None,
    source_service_accounts: List[str] = None,
    denied: List[
        make_dataclass(
            "items",
            [
                ("IPProtocol", str, field(default=None)),
                ("ports", List[str], field(default=None)),
            ],
        ),
    ] = None,
    description: str = None,
    direction: str = None,
    destination_ranges: List[str] = None,
    allowed: List[
        make_dataclass(
            "items",
            [
                ("ports", List[str], field(default=None)),
                ("IPProtocol", str, field(default=None)),
            ],
        ),
    ] = None,
    disabled: bool = None,
    network: str = None,
    priority: int = None,
    log_config: Dict[str, Any] = None,
    target_service_accounts: List[str] = None,
    source_tags: List[str] = None,
    source_ranges: List[str] = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Creates or updates a firewall rule in the specified project using the data included in the request.

    Args:
        name(str): An Idem name of the resource.
        project(str, Optional): Project ID for this request.
        target_tags(List[str], Optional):
            A list of tags that controls which instances the firewall rule applies to. If targetTags are specified,
            then the firewall rule applies only to instances in the VPC network that have one of those tags. If no
            targetTags are specified, the firewall rule applies to all instances on the specified network. Defaults to None.
        source_service_accounts(List[str], Optional):
            If source service accounts are specified, the firewall rules apply only to traffic originating from an
            instance with a service account in this list. Source service accounts cannot be used to control traffic to
            an instance's external IP address because service accounts are associated with an instance, not an IP
            address. sourceRanges can be set at the same time as sourceServiceAccounts. If both are set, the firewall
            applies to traffic that has a source IP address within the sourceRanges OR a source IP that belongs to an
            instance with service account listed in sourceServiceAccount. The connection does not need to match both
            fields for the firewall to apply. sourceServiceAccounts cannot be used at the same time as sourceTags or
            targetTags. Defaults to None.
        denied(List[Dict[str, Any]], Optional):
            The list of DENY rules specified by this firewall. Each rule specifies a protocol and port-range tuple that
            describes a denied connection. Defaults to None.
            * IPProtocol (str, Optional): The IP protocol to which this rule applies. The protocol type is required when
                creating a firewall rule. This value can either be one of the following well known protocol strings
                (tcp, udp, icmp, esp, ah, ipip, sctp) or the IP protocol number.
            * ports (list[str], Optional): An optional list of ports to which this rule applies. This field is only
                applicable for the UDP or TCP protocol. Each entry must be either an integer or a range. If not specified,
                this rule applies to connections through any port. Example inputs include: ["22"], ["80","443"], and
                ["12345-12349"].
        description(str, Optional):
            An optional description of this resource. Provide this field when you create the resource. Defaults to None.
        direction(str, Optional):
            Direction of traffic to which this firewall applies, either `INGRESS` or `EGRESS`. The default is `INGRESS`.
            For `EGRESS` traffic, you cannot specify the sourceTags fields. Defaults to None.
        destination_ranges(List[str], Optional):
            If destination ranges are specified, the firewall rule applies only to traffic that has destination IP
            address in these ranges. These ranges must be expressed in CIDR format. Both IPv4 and IPv6 are supported.
            Defaults to None.
        allowed(List[Dict[str, Any]], Optional):
            The list of ALLOW rules specified by this firewall. Each rule specifies a protocol and port-range tuple that
            describes a permitted connection. Defaults to None.
            * ports (List[str], Optional): An optional list of ports to which this rule applies. This field is only
                applicable for the UDP or TCP protocol. Each entry must be either an integer or a range. If not specified,
                this rule applies to connections through any port. Example inputs include: ["22"], ["80","443"], and ["12345-12349"].
            * IPProtocol (str, Optional): The IP protocol to which this rule applies. The protocol type is required when
                creating a firewall rule. This value can either be one of the following well known protocol strings
                (tcp, udp, icmp, esp, ah, ipip, sctp) or the IP protocol number.
        disabled(bool, Optional):
            Denotes whether the firewall rule is disabled. When set to true, the firewall rule is not enforced and the
            network behaves as if it did not exist. If this is unspecified, the firewall rule will be enabled. Defaults to None.
        network(str, Optional):
            URL of the network resource for this firewall rule. If not specified when creating a firewall rule, the
            default network is used: global/networks/default If you choose to specify this field, you can specify the
            network as a full or partial URL. For example, the following are all valid
            URLs: - https://www.googleapis.com/compute/v1/projects/myproject/global/networks/my-network -
            projects/myproject/global/networks/my-network - global/networks/default . Defaults to None.
        priority(int, Optional):
            Priority for this rule. This is an integer between `0` and `65535`, both inclusive. The default value is
            `1000`. Relative priorities determine which rule takes effect if multiple rules apply. Lower values indicate
            higher priority. For example, a rule with priority `0` has higher precedence than a rule with priority `1`.
            DENY rules take precedence over ALLOW rules if they have equal priority. Note that VPC networks have implied
            rules with a priority of `65535`. To avoid conflicts with the implied rules, use a priority number less than
            `65535`. Defaults to None.
        log_config(Dict[str, Any], Optional):
            This field denotes the logging options for a particular firewall rule. If logging is enabled, logs will be
            exported to Cloud Logging. Defaults to None.
        target_service_accounts(List[str], Optional):
            A list of service accounts indicating sets of instances located in the network that may make network
            connections as specified in allowed[]. targetServiceAccounts cannot be used at the same time as targetTags
            or sourceTags. If neither targetServiceAccounts nor targetTags are specified, the firewall rule applies to
            all instances on the specified network. Defaults to None.
        source_tags(List[str], Optional):
            If source tags are specified, the firewall rule applies only to traffic with source IPs that match the
            primary network interfaces of VM instances that have the tag and are in the same VPC network. Source tags
            cannot be used to control traffic to an instance's external IP address, it only applies to traffic between
            instances in the same virtual network. Because tags are associated with instances, not IP addresses. One or
            both of sourceRanges and sourceTags may be set. If both fields are set, the firewall applies to traffic that
            has a source IP address within sourceRanges OR a source IP from a resource with a matching tag listed in the
            sourceTags field. The connection does not need to match both fields for the firewall to apply. Defaults to None.
        source_ranges(List[str], Optional):
            If source ranges are specified, the firewall rule applies only to traffic that has a source IP address in
            these ranges. These ranges must be expressed in CIDR format. One or both of sourceRanges and sourceTags may
            be set. If both fields are set, the rule applies to traffic that has a source IP address within sourceRanges
            OR a source IP from a resource with a matching tag listed in the sourceTags field. The connection does not
            need to match both fields for the rule to apply. Both IPv4 and IPv6 are supported. Defaults to None.
        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            firewall-present:
              gcp.compute.firewall.present:
              - name: my-firewall-name
              - description: Free text description
              - network: https://www.googleapis.com/compute/v1/projects/project-name/global/networks/default
              - source_ranges:
                - 0.0.0.0/0
              - allowed:
                - ip_protocol: tcp
                  ports:
                  - '22'
              - direction: INGRESS
              - log_config:
                  enable: false
              - disabled: false
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)
    resource_type_camel = hub.tool.gcp.case.camel(RESOURCE_TYPE_FULL.split(".")[-1])

    if hub.tool.gcp.resource_prop_utils.properties_mismatch_resource_id(
        RESOURCE_TYPE, resource_id, {**locals(), resource_type_camel: name}
    ):
        result["comment"].append(
            hub.tool.gcp.comment_utils.properties_mismatch_resource_id_comment(
                RESOURCE_TYPE_FULL, name
            )
        )

    if ctx.get("rerun_data"):
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, ctx.get("rerun_data"), RESOURCE_TYPE
        )

        if not handle_operation_ret["result"]:
            result["comment"] += handle_operation_ret["comment"]
            if handle_operation_ret.get("rerun_data"):
                result["rerun_data"] = handle_operation_ret["rerun_data"]
                if handle_operation_ret["rerun_data"].get("has_error", False):
                    result["result"] = False
            else:
                result["result"] = False

            return result

        resource_id = handle_operation_ret["resource_id"]

    get_resource_only_with_resource_id = hub.OPT.idem.get(
        "get_resource_only_with_resource_id", False
    )
    if resource_id:
        old_get_ret = await hub.exec.gcp.compute.firewall.get(
            ctx, resource_id=resource_id
        )

        if not old_get_ret["result"] or (
            not old_get_ret["ret"] and get_resource_only_with_resource_id
        ):
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            return result

        # long running operation has succeeded - both update and create
        if ctx.get("rerun_data"):
            result["new_state"] = old_get_ret["ret"]
            result["old_state"] = ctx["rerun_data"]["old_state"]
            if ctx["rerun_data"]["old_state"]:
                result["comment"].append(
                    hub.tool.gcp.comment_utils.update_comment(RESOURCE_TYPE_FULL, name)
                )
            else:
                result["comment"].append(
                    hub.tool.gcp.comment_utils.create_comment(RESOURCE_TYPE_FULL, name)
                )
            return result

        result["old_state"] = old_get_ret["ret"]
    elif not get_resource_only_with_resource_id:
        resource_id = hub.tool.gcp.resource_prop_utils.construct_resource_id(
            RESOURCE_TYPE, {"project": project, "firewall": name}
        )
        old_get_ret = await hub.exec.gcp.compute.firewall.get(
            ctx, resource_id=resource_id
        )

        if not old_get_ret["result"]:
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            return result

        if old_get_ret["ret"]:
            result["old_state"] = old_get_ret["ret"]

    resource_body = {
        "priority": priority,
        "log_config": log_config,
        "direction": direction,
        "source_ranges": source_ranges,
        "source_service_accounts": source_service_accounts,
        "allowed": allowed,
        "target_tags": target_tags,
        "source_tags": source_tags,
        "denied": denied,
        "target_service_accounts": target_service_accounts,
        "network": network,
        "destination_ranges": destination_ranges,
        "disabled": disabled,
        "name": name,
        "description": description,
    }

    resource_body = {k: v for (k, v) in resource_body.items() if v is not None}
    operation = None
    if result["old_state"]:
        changes = hub.tool.gcp.utils.compare_states(
            result["old_state"], resource_body, RESOURCE_TYPE
        )
        if not changes:
            result["result"] = True
            result["comment"].append(
                hub.tool.gcp.comment_utils.up_to_date_comment(RESOURCE_TYPE_FULL, name)
            )
            result["new_state"] = result["old_state"]
            return result

        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_update_comment(
                    RESOURCE_TYPE_FULL, name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                {"resource_id": resource_id, **resource_body}
            )
            return result

        update_ret = await hub.exec.gcp_api.client.compute.firewall.patch(
            hub, ctx, resource_id=resource_id, body=resource_body
        )
        if not update_ret["result"] or not update_ret["ret"]:
            result["result"] = False
            result["comment"] += update_ret["comment"]
            return result

        if hub.tool.gcp.operation_utils.is_operation(update_ret.get("ret")):
            operation = update_ret["ret"]

        get_ret = await hub.exec.gcp.compute.firewall.get(ctx, resource_id=resource_id)
        if not get_ret["result"] and not get_ret["ret"]:
            result["result"] = False
            result["comment"] += get_ret["comment"]
            return result

        result["new_state"] = get_ret["ret"]
        result["comment"].append(
            hub.tool.gcp.comment_utils.update_comment(RESOURCE_TYPE_FULL, name)
        )
        return result
    else:
        if ctx.get("test", False):
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    RESOURCE_TYPE_FULL, name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                {"resource_id": resource_id, **resource_body}
            )
            return result
        create_ret = await hub.exec.gcp_api.client.compute.firewall.insert(
            ctx, project=project, body=resource_body
        )
        if not create_ret["result"] or not create_ret["ret"]:
            result["result"] = False
            if create_ret["comment"] and next(
                (
                    comment
                    for comment in create_ret["comment"]
                    if "alreadyExists" in comment
                ),
                None,
            ):
                result["comment"].append(
                    hub.tool.gcp.comment_utils.already_exists_comment(
                        RESOURCE_TYPE_FULL, name
                    )
                )
            else:
                result["comment"] += create_ret["comment"]
            return result

        if hub.tool.gcp.operation_utils.is_operation(create_ret.get("ret")):
            operation = create_ret["ret"]

    if operation:
        operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
            operation.get("selfLink"), "compute.global_operation"
        )
        result["rerun_data"] = {
            "operation_id": operation_id,
            "old_state": result["old_state"],
        }
        result["result"] = False
        return result

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves the list of firewall rules available to the specified project.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.firewall
    """
    result = {}

    describe_ret = await hub.exec.gcp.compute.firewall.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(
            f"Could not describe gcp.compute.firewall {describe_ret['comment']}"
        )
        return result

    for resource in describe_ret["ret"]:
        resource_id = resource.get("resource_id")

        result[resource_id] = {
            "gcp.compute.firewall.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



"""State module for managing Disks."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from dict_tools.typing import Computed

from idem_gcp.tool.gcp.utils import zonal_absent

# prevent commit hook from removing the import
absent = zonal_absent

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str = None,
    project: str = None,
    zone: str = None,
    region: str = None,
    resource_id: str = None,
    request_id: str = None,
    source_image: str = None,
    size_gb: str = None,
    source_snapshot: str = None,
    source_storage_object: str = None,
    options: str = None,
    type_: (str, "alias=type") = None,
    licenses: List[str] = None,
    guest_os_features: List[
        make_dataclass(
            "GuestOsFeature", [("type_", (str, "alias=type"), field(default=None))]
        )
    ] = None,
    disk_encryption_key: make_dataclass(
        "CustomerEncryptionKey",
        [
            ("raw_key", str, field(default=None)),
            ("rsa_encrypted_key", str, field(default=None)),
            ("kms_key_name", str, field(default=None)),
            ("kms_key_service_account", str, field(default=None)),
        ],
    ) = None,
    source_image_encryption_key: make_dataclass(
        "CustomerEncryptionKey",
        [
            ("raw_key", str, field(default=None)),
            ("rsa_encrypted_key", str, field(default=None)),
            ("kms_key_name", str, field(default=None)),
            ("kms_key_service_account", str, field(default=None)),
        ],
    ) = None,
    source_snapshot_encryption_key: make_dataclass(
        "CustomerEncryptionKey",
        [
            ("raw_key", str, field(default=None)),
            ("rsa_encrypted_key", str, field(default=None)),
            ("kms_key_name", str, field(default=None)),
            ("kms_key_service_account", str, field(default=None)),
        ],
    ) = None,
    labels: Dict[str, str] = None,
    label_fingerprint: str = None,
    replica_zones: List[str] = None,
    license_codes: List[str] = None,
    physical_block_size_bytes: str = None,
    resource_policies: List[str] = None,
    source_disk: str = None,
    provisioned_iops: str = None,
    location_hint: str = None,
    architecture: str = None,
    params: make_dataclass(
        "DiskParams",
        [
            ("resource_manager_tags", Dict[str, str], field(default=None)),
        ],
    ) = None,
    description: str = None,
    id_: (Computed[str], "alias=id") = None,
) -> Dict[str, Any]:
    r"""Creates or updates a persistent disk in the specified project using the data in the request.

    You can create a disk from a source (sourceImage, sourceSnapshot, or sourceDisk) or create an empty 500 GB data disk by omitting all properties.
    You can also create a disk that is larger than the default size by specifying the sizeGb property.

    Args:
        name(str):
            An Idem name of the resource.

        zone(str, Optional):
            The name of the zone for this request.

        region(str, Optional):
            The name of the region for this request.

        project(str, Optional):
            Project ID for this request.

        resource_id(str, Optional):
            An identifier of the resource in the provider.
            Defaults to None.

        request_id(str, Optional):
            An optional request ID to identify requests.
            Specify a unique request ID so that if you must retry your request,
            the server will know to ignore the request if it has already been completed.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID,
            the server can check if original operation with the same request ID was received, and if so,
            will ignore the second request. This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that
            zero UUID is not supported ( 00000000-0000-0000-0000-000000000000).
            Defaults to None.

        source_image(str, Optional):
            The source image used to create this disk. If the source image is deleted, this field will not be set.
            To create a disk with one of the public operating system images, specify the image by its family name.
            For example, specify family/debian-9 to use the latest Debian 9 image:
            projects/debian-cloud/global/images/family/debian-9
            Alternatively, use a specific version of a public operating system image:
            projects/debian-cloud/global/images/debian-9-stretch-vYYYYMMDD
            To create a disk with a custom image that you created, specify the image name in the following format:
            global/images/my-custom-image You can also specify a custom image by its image family,
            which returns the latest version of the image in that family.
            Replace the image name with family/family-name: global/images/family/my-image-family .
            Defaults to None.

        size_gb(str, Optional):
            Size, in GB, of the persistent disk.
            You can specify this field when creating a persistent disk using the sourceImage, sourceSnapshot,
            or sourceDisk parameter, or specify it alone to create an empty persistent disk.
            If you specify this field along with a source, the value of sizeGb must not
            be less than the size of the source. Acceptable values are 1 to 65536, inclusive.
            Defaults to None.

        source_snapshot(str, Optional):
            The source snapshot used to create this disk. You can provide this as a partial or full URL to the resource.
            For example, the following are valid values:
                - https://www.googleapis.com/compute/v1/projects/project /global/snapshots/snapshot
                - projects/project/global/snapshots/snapshot
                - global/snapshots/snapshot
            Defaults to None.

        source_storage_object(str, Optional):
            The full Google Cloud Storage URI where the disk image is stored.
            This file must be a gzip-compressed tarball whose name ends in .tar.gz or virtual machine disk
            whose name ends in vmdk. Valid URIs may start with gs:// or https://storage.googleapis.com/.
            This flag is not optimized for creating multiple disks from a source storage object.
            To create many disks from a source storage object, use gcloud compute images import instead.
            Defaults to None.

        options(str, Optional):
            Internal use only.
            Defaults to None.

        type(str, Optional):
            URL of the disk type resource describing which disk type to use to create the disk.
            Provide this when creating the disk. For example: projects/project /zones/zone/diskTypes/pd-ssd.
            See Persistent disk types.
            Defaults to None.

        licenses(List[str], Optional):
            A list of publicly visible licenses. Reserved for Google's use.
            Defaults to None.

        guest_os_features(List[Dict[str, Any]], Optional):
            A list of features to enable on the guest operating system. Applicable only for bootable images.
            Read Enabling guest operating system features to see a list of available options.
            Defaults to None.

                * type_ (str, Optional):
                    The ID of a supported feature. To add multiple values,
                    use commas to separate values. Set to one or more of the following values:
                        - VIRTIO_SCSI_MULTIQUEUE
                        - WINDOWS
                        - MULTI_IP_SUBNET
                        - UEFI_COMPATIBLE
                        - GVNIC
                        - SEV_CAPABLE
                        - SUSPEND_RESUME_COMPATIBLE
                        - SEV_SNP_CAPABLE
                    For more information, see Enabling guest operating system features.

        disk_encryption_key(Dict[str, Any], Optional):
            Encrypts the disk using a customer-supplied encryption key or a customer-managed encryption key.
            Encryption keys do not protect access to metadata of the disk. After you encrypt a disk with a
            customer-supplied key, you must provide the same key if you use the disk later. For example,
            to create a disk snapshot, to create a disk image, to create a machine image, or to attach the disk to a
            virtual machine. After you encrypt a disk with a customer-managed key, the diskEncryptionKey.kmsKeyName
            is set to a key *version* name once the disk is created. The disk is encrypted with this version of the key.
            In the response, diskEncryptionKey.kmsKeyName appears in the following format:
            "diskEncryptionKey.kmsKeyName":
            "projects/kms_project_id/locations/region/keyRings/ key_region/cryptoKeys/key /cryptoKeysVersions/version"
            If you do not provide an encryption key when creating the disk, then the disk is encrypted using an
            automatically generated key and you don't need to provide a key to use the disk later.
            Defaults to None.

                * raw_key(str, Optional):
                    Specifies a 256-bit customer-supplied encryption key, encoded in RFC 4648 base64 to either encrypt
                    or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey.
                    For example: \"raw_key\": \"SGVsbG8gZnJvbSBHb29nbGUgQ2xvdWQgUGxhdGZvcm0=\"

                * rsa_encrypted_key(str, Optional):
                    Specifies an RFC 4648 base64 encoded, RSA-wrapped 2048-bit customer-supplied encryption key to
                    either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey.
                    For example:
                        \"rsa_encrypted_key\":
                        \"ieCx/NcW06PcT7Ep1X6LUTc/hLvUDYyzSZPPVCVPTVEohpeHASqC8uw5TzyO9U+Fka9JFH z0mBibXUInrC/jEk014kCK/
                        NPjYgEMOyssZ4ZINPKxlUh2zn1bV+MCaTICrdmuSBTWlUUiFoD D6PYznLwh8ZNdaheCeZ8ewEXgFQ8V+sDroLaN3Xs3MDTXQEM
                        MoNUXMCZEIpg9Vtp9x2oe==\"
                    The key must meet the following requirements before you can provide it to Compute Engine:
                        1. The key is wrapped using a RSA public key certificate provided by Google.
                        2. After being wrapped, the key must be encoded in RFC 4648 base64 encoding.
                    Gets the RSA public key certificate provided by Google at:
                    https://cloud-certs.storage.googleapis.com/google-cloud-csek-ingress.pem

                * kms_key_name(str, Optional):
                    The name of the encryption key that is stored in Google Cloud KMS.
                    For example:
                        \"kms_key_name\":
                        \"projects/kms_project_id/locations/region/keyRings/ key_region/cryptoKeys/key

                * kms_key_service_account(str, Optional):
                    The service account being used for the encryption request for the given KMS key. If absent,
                    the Compute Engine default service account is used. For example:
                        \"kms_key_service_account\":
                        \"name@project_id.iam.gserviceaccount.com/


        source_image_encryption_key(Dict[str, Any], Optional):
            The customer-supplied encryption key of the source image.
            Required if the source image is protected by a customer-supplied encryption key.
            Defaults to None.

                * raw_key(str, Optional)
                * rsa_encrypted_key(str, Optional)
                * kms_key_name(str, Optional)
                * kms_key_service_account(str, Optional)

        source_snapshot_encryption_key(Dict[str, Any], Optional):
            The customer-supplied encryption key of the source snapshot.
            Required if the source snapshot is protected by a customer-supplied encryption key.
            Defaults to None.

                * raw_key(str, Optional)
                * rsa_encrypted_key(str, Optional)
                * kms_key_name(str, Optional)
                * kms_key_service_account(str, Optional)

        labels(Dict[str, Dict[str, str]], Optional):
            Labels to apply to this disk. These can be later modified by the setLabels method.
            Defaults to None.

        label_fingerprint(str, Optional):
            A fingerprint for the labels being applied to this disk,
            which is essentially a hash of the labels set used for optimistic locking.
            The fingerprint is initially generated by Compute Engine and
            changes after every request to modify or update labels.
            You must always provide an up-to-date fingerprint hash in order to update or change labels,
            otherwise the request will fail with error 412 conditionNotMet.
            To see the latest fingerprint, make a get() request to retrieve a disk.
            Defaults to None.

        replica_zones(List[str], Optional):
            URLs of the zones where the disk should be replicated to. Only applicable for regional resources.
            Defaults to None.

        license_codes(List[str], Optional):
            Integer license codes indicating which licenses are attached to this disk.
            Defaults to None.

        physical_block_size_bytes(str, Optional):
            Physical block size of the persistent disk, in bytes. If not present in a request, a default value is used.
            The currently supported size is 4096, other sizes may be added in the future.
            If an unsupported value is requested,
            the error message will list the supported values for the caller's project.
            Defaults to None.

        resource_policies(List[str], Optional):
            Resource policies applied to this disk for automatic snapshot creations.
            To remove a policy set resource_policies to be an empty array e.g. []
            Defaults to None.

        source_disk(str, Optional):
            The source disk used to create this disk. You can provide this as a partial or full URL to the resource.
            For example, the following are valid values:
                - https://www.googleapis.com/compute/v1/projects/project/zones/zone /disks/disk
                - https://www.googleapis.com/compute/v1/projects/project/regions/region /disks/disk
                - projects/project/zones/zone/disks/disk
                - projects/project/regions/region/disks/disk
                - zones/zone/disks/disk
                - regions/region/disks/disk
                Defaults to None.

        provisioned_iops(str, Optional):
            Indicates how many IOPS to provision for the disk.
            This sets the number of I/O operations per second that the disk can handle.
            Values must be between 10,000 and 120,000. For more details, see the Extreme persistent disk documentation.
            Defaults to None.

        location_hint(str, Optional):
            An opaque location hint used to place the disk close to other resources.
            This field is for use by internal tools that use the public API.
            Defaults to None.

        architecture(str, Optional):
            The architecture of the disk. Valid values are ARM64 or X86_64.
            Defaults to None.

        params(Dict[str, Dict[str, str]], Optional):
            Input only. [Input Only] Additional params passed with the request,
            but not persisted as part of resource payload.
            Defaults to None.

                * resourceManagerTags(Dict[str, str], Optional):
                    Resource manager tags to be bound to the disk. Tag keys and values have the same definition as
                    resource manager tags. Keys must be in the format `tagKeys/{tag_key_id}`, and values are in the format
                    `tagValues/456`. The field is ignored (both PUT & PATCH) when empty.

        description(str, Optional):
            An optional description of this resource.
            Defaults to None.

        id(str, optional): The unique identifier for the resource. This identifier is defined by the server. Read-only property

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            example_resource_name:
              gcp.compute.disk.present:
                - project: project-name
                - zone: us-central1-a
                - description: Description for example_resource_name
                - type: projects/project-name/zones/us-central1-a/diskTypes/pd-balanced
                - size_gb: 10
                - resource_policies:
                    - projects/project-name/regions/us-central1/resourcePolicies/default-schedule-1
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if ctx.get("wrapper_result"):
        result = ctx.get("wrapper_result")

    # to be autogenerated by pop-create based on insert/update props in properties.yaml
    resource_body = {
        "type_": type_,
        "name": name,
        "source_image": source_image,
        "label_fingerprint": label_fingerprint,
        "architecture": architecture,
        "guest_os_features": guest_os_features,
        "source_disk": source_disk,
        "resource_policies": resource_policies,
        "options": options,
        "source_storage_object": source_storage_object,
        "licenses": licenses,
        "size_gb": size_gb,
        "license_codes": license_codes,
        "source_snapshot_encryption_key": source_snapshot_encryption_key,
        "zone": zone,
        "region": region,
        "source_image_encryption_key": source_image_encryption_key,
        "physical_block_size_bytes": physical_block_size_bytes,
        "source_snapshot": source_snapshot,
        "description": description,
        "labels": labels,
        "params": params,
        "replica_zones": replica_zones,
        "location_hint": location_hint,
        "disk_encryption_key": disk_encryption_key,
        "provisioned_iops": provisioned_iops,
        "id_": id_,
    }

    resource_body = {k: v for (k, v) in resource_body.items() if v is not None}
    operation = None
    if result["old_state"]:
        # Update
        resource_id = result["old_state"].get("resource_id", None)
        zone = hub.tool.gcp.resource_prop_utils.parse_link_to_zone(
            result["old_state"]["zone"]
        )
        result["old_state"]["zone"] = zone

        changes = hub.tool.gcp.utils.compare_states(
            result["old_state"],
            {
                "resource_id": resource_id,
                **resource_body,
            },
            "compute.disk",
        )

        resource_updated = False
        if changes:
            changed_non_updatable_properties = (
                hub.tool.gcp.resource_prop_utils.get_changed_non_updatable_properties(
                    "compute.disk", changes
                )
            )
            if changed_non_updatable_properties:
                result["result"] = False
                result["comment"].append(
                    hub.tool.gcp.comment_utils.non_updatable_properties_comment(
                        "gcp.compute.disk",
                        name,
                        changed_non_updatable_properties,
                    )
                )
                result["new_state"] = result["old_state"]
                return result

            # Resize
            size_gb_updated = False
            if size_gb is not None and str(result["old_state"]["size_gb"]) != str(
                size_gb
            ):
                if not ctx.get("test"):
                    resize_body = {"size_gb": size_gb}
                    update_ret = await hub.exec.gcp_api.client.compute.disk.resize(
                        ctx,
                        resource_id=resource_id,
                        body=resize_body,
                        request_id=request_id,
                    )

                    if not update_ret["result"]:
                        result["result"] = False
                        result["comment"] += update_ret["comment"]
                        return result

                    if hub.tool.gcp.operation_utils.is_operation(update_ret["ret"]):
                        operation = update_ret["ret"]
                        operation_type = (
                            hub.tool.gcp.operation_utils.get_operation_type(
                                operation.get("selfLink")
                            )
                        )
                        operation_id = (
                            hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                                operation.get("selfLink"), operation_type
                            )
                        )
                        handle_operation_ret = (
                            await hub.tool.gcp.operation_utils.handle_operation(
                                ctx,
                                {"operation_id": operation_id},
                                "compute.disk",
                                True,
                            )
                        )
                        if not handle_operation_ret["result"]:
                            result["result"] = False
                            result["comment"] += handle_operation_ret["comment"]
                            return result

                size_gb_updated = True

            # Update labels
            labels_updated = False
            if (
                labels is not None
                and not (labels == {} and result["old_state"].get("labels") is None)
                and result["old_state"].get("labels") != labels
            ):
                if not ctx.get("test"):
                    set_labels_body = {
                        "labels": labels,
                        "label_fingerprint": label_fingerprint,
                    }
                    kwargs = {
                        "project": project,
                        "resource": name,
                        "body": set_labels_body,
                        "request_id": request_id,
                    }
                    if zone:
                        kwargs.update({"zone": zone})
                    if region:
                        kwargs.update({"region": region})

                    update_ret = await hub.exec.gcp_api.client.compute.disk.setLabels(
                        ctx,
                        **kwargs,
                    )

                    if not update_ret["result"]:
                        result["result"] = False
                        result["comment"] += update_ret["comment"]
                        return result

                    if hub.tool.gcp.operation_utils.is_operation(update_ret["ret"]):
                        operation = update_ret["ret"]
                        operation_type = (
                            hub.tool.gcp.operation_utils.get_operation_type(
                                operation.get("selfLink")
                            )
                        )
                        operation_id = (
                            hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                                operation.get("selfLink"), operation_type
                            )
                        )
                        handle_operation_ret = (
                            await hub.tool.gcp.operation_utils.handle_operation(
                                ctx,
                                {"operation_id": operation_id},
                                "compute.disk",
                                True,
                            )
                        )
                        if not handle_operation_ret["result"]:
                            result["result"] = False
                            result["comment"] += handle_operation_ret["comment"]
                            return result

                labels_updated = True

            # Update resource policies
            resource_policies_updated = False
            if not hub.tool.gcp.state_comparison_utils.are_lists_identical(
                result["old_state"].get("resource_policies"), resource_policies
            ):
                # If (some) resource policies are links, parse them to resource policy resource ids
                new_resource_policies = [
                    hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                        policy, "compute.resource_policy"
                    )
                    for policy in (
                        resource_policies if resource_policies is not None else []
                    )
                ]
                current_resource_policies = [
                    hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                        policy, "compute.resource_policy"
                    )
                    for policy in result["old_state"].get("resource_policies", [])
                ]

                resource_policies_to_remove = [
                    policy
                    for policy in current_resource_policies
                    if policy not in new_resource_policies
                ]
                resource_policies_to_add = [
                    policy
                    for policy in new_resource_policies
                    if policy not in current_resource_policies
                ]

                if (
                    len(resource_policies_to_remove) > 0
                    or len(resource_policies_to_add) > 0
                ):
                    if not ctx.get("test"):
                        add_request_body = None
                        if len(resource_policies_to_add) > 0:
                            add_request_body = {
                                "resource_policies": resource_policies_to_add
                            }
                        remove_request_body = None
                        if len(resource_policies_to_remove) > 0:
                            remove_request_body = {
                                "resource_policies": resource_policies_to_remove
                            }

                        remove_ret = None
                        if remove_request_body is not None:
                            remove_ret = await hub.exec.gcp_api.client.compute.disk.removeResourcePolicies(
                                ctx,
                                resource_id=resource_id,
                                body=remove_request_body,
                                request_id=request_id,
                            )

                            if not remove_ret["result"]:
                                result["result"] = False
                                result["comment"] += remove_ret["comment"]
                                return result

                            if hub.tool.gcp.operation_utils.is_operation(
                                remove_ret["ret"]
                            ):
                                operation = remove_ret["ret"]
                                operation_type = (
                                    hub.tool.gcp.operation_utils.get_operation_type(
                                        operation.get("selfLink")
                                    )
                                )
                                operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                                    operation.get("selfLink"), operation_type
                                )
                                handle_operation_ret = (
                                    await hub.tool.gcp.operation_utils.handle_operation(
                                        ctx,
                                        {"operation_id": operation_id},
                                        "compute.disk",
                                        True,
                                    )
                                )
                                if not handle_operation_ret["result"]:
                                    result["result"] = False
                                    result["comment"] += handle_operation_ret["comment"]
                                    return result

                        if (
                            remove_ret is None or remove_ret["result"]
                        ) and add_request_body is not None:
                            add_ret = await hub.exec.gcp_api.client.compute.disk.addResourcePolicies(
                                ctx,
                                resource_id=resource_id,
                                body=add_request_body,
                                request_id=request_id,
                            )

                            if not add_ret["result"]:
                                result["result"] = False
                                result["comment"] += add_ret["comment"]
                                return result

                            if hub.tool.gcp.operation_utils.is_operation(
                                add_ret["ret"]
                            ):
                                operation = add_ret["ret"]
                                operation_type = (
                                    hub.tool.gcp.operation_utils.get_operation_type(
                                        operation.get("selfLink")
                                    )
                                )
                                operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                                    operation.get("selfLink"), operation_type
                                )
                                handle_operation_ret = (
                                    await hub.tool.gcp.operation_utils.handle_operation(
                                        ctx,
                                        {"operation_id": operation_id},
                                        "compute.disk",
                                        True,
                                    )
                                )
                                if not handle_operation_ret["result"]:
                                    result["comment"] += handle_operation_ret["comment"]
                                    return result

                    resource_policies_updated = True

            resource_updated = (
                size_gb_updated or labels_updated or resource_policies_updated
            )

        # Reset the operation because we don't need it
        # since we are waiting for the operations to finish in order to proceed
        operation = None
        if resource_updated:
            if ctx.get("test", False):
                result["comment"].append(
                    hub.tool.gcp.comment_utils.would_update_comment(
                        "gcp.compute.disk", name
                    )
                )
                result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                    {
                        "resource_id": resource_id,
                        **result["old_state"],
                        **resource_body,
                    }
                )
                result["new_state"]["resource_id"] = resource_id
            else:
                get_ret = await hub.exec.gcp.compute.disk.get(
                    ctx, name=name, resource_id=resource_id
                )
                if not get_ret["result"] and not get_ret["ret"]:
                    result["result"] = False
                    result["comment"] += get_ret["comment"]
                    return result

                result["new_state"] = get_ret["ret"]
                result["comment"].append(
                    hub.tool.gcp.comment_utils.update_comment("gcp.compute.disk", name)
                )
                return result
        else:
            result["new_state"] = result["old_state"]
            result["comment"].append(
                hub.tool.gcp.comment_utils.up_to_date_comment("gcp.compute.disk", name)
            )
    else:
        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    "gcp.compute.disk", name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                {"resource_id": resource_id, **resource_body}
            )
            result["new_state"]["resource_id"] = resource_id
            return result

        # Create
        kwargs = {"name": name, "project": project, "body": resource_body}
        if zone:
            kwargs.update({"zone": zone})
        if region:
            kwargs.update({"region": region})

        create_ret = await hub.exec.gcp_api.client.compute.disk.insert(ctx, **kwargs)
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] += create_ret["comment"]
        else:
            result["comment"].append(
                hub.tool.gcp.comment_utils.create_comment("gcp.compute.disk", name)
            )
            resource_id = create_ret["ret"].get("resource_id")
        if create_ret["ret"] is not None:
            if hub.tool.gcp.operation_utils.is_operation(create_ret["ret"]):
                operation = create_ret["ret"]

    if operation:
        operation_type = hub.tool.gcp.operation_utils.get_operation_type(
            operation.get("selfLink")
        )
        operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
            operation.get("selfLink"), operation_type
        )
        result["rerun_data"] = {
            "operation_id": operation_id,
            "old_state": result["old_state"],
        }
        return result

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves a list of persistent disks contained within the specified zone.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.disk
    """
    result = {}

    describe_ret = await hub.exec.gcp.compute.disk.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(f"Could not describe disks {describe_ret['comment']}")
        return {}

    for resource in describe_ret["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "gcp.compute.disk.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)
